import pathlib
import sys
from typing import Callable, List, Optional, Sequence, Union
from unittest.mock import patch

import numpy as np
import pyarrow as pa
import pytest
import tiledbsoma as soma
from scipy import sparse
from scipy.sparse import coo_matrix, spmatrix
from somacore import AxisQuery
from tiledbsoma import Experiment, _factory
from tiledbsoma._collection import CollectionBase
from torch.utils.data._utils.worker import WorkerInfo

# conditionally import torch, as it will not be available in all test environments
try:
    from torch import Tensor, float32

    from cellxgene_census.experimental.ml.pytorch import (
        ExperimentDataPipe,
        experiment_dataloader,
    )
except ImportError:
    # this should only occur when not running `experimental`-marked tests
    pass


def pytorch_x_value_gen(obs_range: range, var_range: range) -> spmatrix:
    occupied_shape = (obs_range.stop - obs_range.start, var_range.stop - var_range.start)
    checkerboard_of_ones = coo_matrix(np.indices(occupied_shape).sum(axis=0) % 2)
    checkerboard_of_ones.row += obs_range.start
    checkerboard_of_ones.col += var_range.start
    return checkerboard_of_ones


def pytorch_seq_x_value_gen(obs_range: range, var_range: range) -> spmatrix:
    """A sparse matrix where the values of each col are the obs_range values. Useful for checking the
    X values are being returned in the correct order."""
    data = np.vstack([list(obs_range)] * len(var_range)).flatten()
    rows = np.vstack([list(obs_range)] * len(var_range)).flatten()
    cols = np.column_stack([list(var_range)] * len(obs_range)).flatten()
    return coo_matrix((data, (rows, cols)))


@pytest.fixture
def X_layer_names() -> List[str]:
    return ["raw"]


@pytest.fixture
def obsp_layer_names() -> Optional[List[str]]:
    return None


@pytest.fixture
def varp_layer_names() -> Optional[List[str]]:
    return None


def add_dataframe(coll: CollectionBase, key: str, value_range: range) -> None:
    df = coll.add_new_dataframe(
        key,
        schema=pa.schema(
            [
                ("soma_joinid", pa.int64()),
                ("label", pa.large_string()),
            ]
        ),
        index_column_names=["soma_joinid"],
    )
    df.write(
        pa.Table.from_pydict(
            {
                "soma_joinid": list(value_range),
                "label": [str(i) for i in value_range],
            }
        )
    )


def add_sparse_array(
    coll: CollectionBase,
    key: str,
    obs_range: range,
    var_range: range,
    value_gen: Callable[[range, range], spmatrix],
) -> None:
    a = coll.add_new_sparse_ndarray(key, type=pa.float32(), shape=(obs_range.stop, var_range.stop))
    tensor = pa.SparseCOOTensor.from_scipy(value_gen(obs_range, var_range))
    a.write(tensor)


@pytest.fixture(scope="function")
def soma_experiment(
    tmp_path: pathlib.Path,
    obs_range: Union[int, range],
    var_range: Union[int, range],
    X_value_gen: Callable[[range, range], sparse.spmatrix],
    obsp_layer_names: Sequence[str],
    varp_layer_names: Sequence[str],
) -> soma.Experiment:
    with soma.Experiment.create((tmp_path / "exp").as_posix()) as exp:
        if isinstance(obs_range, int):
            obs_range = range(obs_range)
        if isinstance(var_range, int):
            var_range = range(var_range)

        add_dataframe(exp, "obs", obs_range)
        ms = exp.add_new_collection("ms")
        rna = ms.add_new_collection("RNA", soma.Measurement)
        add_dataframe(rna, "var", var_range)
        rna_x = rna.add_new_collection("X", soma.Collection)
        add_sparse_array(rna_x, "raw", obs_range, var_range, X_value_gen)

        if obsp_layer_names:
            obsp = rna.add_new_collection("obsp")
            for obsp_layer_name in obsp_layer_names:
                add_sparse_array(obsp, obsp_layer_name, obs_range, var_range, X_value_gen)

        if varp_layer_names:
            varp = rna.add_new_collection("varp")
            for varp_layer_name in varp_layer_names:
                add_sparse_array(varp, varp_layer_name, obs_range, var_range, X_value_gen)
    return _factory.open((tmp_path / "exp").as_posix())


@pytest.mark.experimental
# noinspection PyTestParametrized
@pytest.mark.parametrize(
    "obs_range,var_range,X_value_gen,use_eager_fetch",
    [(6, 3, pytorch_x_value_gen, use_eager_fetch) for use_eager_fetch in (True, False)],
)
def test_non_batched(soma_experiment: Experiment, use_eager_fetch: bool) -> None:
    exp_data_pipe = ExperimentDataPipe(
        soma_experiment,
        measurement_name="RNA",
        X_name="raw",
        obs_column_names=["label"],
        use_eager_fetch=use_eager_fetch,
    )
    row_iter = iter(exp_data_pipe)

    row = next(row_iter)
    assert row[0].int().tolist() == [0, 1, 0]
    assert row[1].tolist() == [0, 0]


@pytest.mark.experimental
# noinspection PyTestParametrized,DuplicatedCode
@pytest.mark.parametrize(
    "obs_range,var_range,X_value_gen,use_eager_fetch",
    [(6, 3, pytorch_x_value_gen, use_eager_fetch) for use_eager_fetch in (True, False)],
)
def test_batching__all_batches_full_size(soma_experiment: Experiment, use_eager_fetch: bool) -> None:
    exp_data_pipe = ExperimentDataPipe(
        soma_experiment,
        measurement_name="RNA",
        X_name="raw",
        obs_column_names=["label"],
        batch_size=3,
        use_eager_fetch=use_eager_fetch,
    )
    batch_iter = iter(exp_data_pipe)

    batch = next(batch_iter)
    assert batch[0].int().tolist() == [[0, 1, 0], [1, 0, 1], [0, 1, 0]]
    assert batch[1].tolist() == [[0, 0], [1, 1], [2, 2]]

    batch = next(batch_iter)
    assert batch[0].int().tolist() == [[1, 0, 1], [0, 1, 0], [1, 0, 1]]
    assert batch[1].tolist() == [[3, 3], [4, 4], [5, 5]]

    with pytest.raises(StopIteration):
        next(batch_iter)


@pytest.mark.experimental
# noinspection PyTestParametrized,DuplicatedCode
@pytest.mark.parametrize(
    "obs_range,var_range,X_value_gen,use_eager_fetch",
    [(range(100_000_000, 100_000_003), 3, pytorch_x_value_gen, use_eager_fetch) for use_eager_fetch in (True, False)],
)
def test_unique_soma_joinids(soma_experiment: Experiment, use_eager_fetch: bool) -> None:
    exp_data_pipe = ExperimentDataPipe(
        soma_experiment,
        measurement_name="RNA",
        X_name="raw",
        obs_column_names=["label"],
        batch_size=3,
        use_eager_fetch=use_eager_fetch,
    )

    soma_joinids = np.concatenate([batch[1][:, 0].numpy() for batch in exp_data_pipe])

    assert len(np.unique(soma_joinids)) == len(soma_joinids)


@pytest.mark.experimental
# noinspection PyTestParametrized
@pytest.mark.parametrize(
    "obs_range,var_range,X_value_gen,use_eager_fetch",
    [(5, 3, pytorch_x_value_gen, use_eager_fetch) for use_eager_fetch in (True, False)],
)
def test_batching__partial_final_batch_size(soma_experiment: Experiment, use_eager_fetch: bool) -> None:
    exp_data_pipe = ExperimentDataPipe(
        soma_experiment,
        measurement_name="RNA",
        X_name="raw",
        obs_column_names=["label"],
        batch_size=3,
        use_eager_fetch=use_eager_fetch,
    )
    batch_iter = iter(exp_data_pipe)

    next(batch_iter)
    batch = next(batch_iter)
    assert batch[0].int().tolist() == [[1, 0, 1], [0, 1, 0]]

    with pytest.raises(StopIteration):
        next(batch_iter)


@pytest.mark.experimental
# noinspection PyTestParametrized,DuplicatedCode
@pytest.mark.parametrize(
    "obs_range,var_range,X_value_gen,use_eager_fetch",
    [(3, 3, pytorch_x_value_gen, use_eager_fetch) for use_eager_fetch in (True, False)],
)
def test_batching__exactly_one_batch(soma_experiment: Experiment, use_eager_fetch: bool) -> None:
    exp_data_pipe = ExperimentDataPipe(
        soma_experiment,
        measurement_name="RNA",
        X_name="raw",
        obs_column_names=["label"],
        batch_size=3,
        use_eager_fetch=use_eager_fetch,
    )
    batch_iter = iter(exp_data_pipe)

    batch = next(batch_iter)
    assert batch[0].int().tolist() == [[0, 1, 0], [1, 0, 1], [0, 1, 0]]
    assert batch[1].tolist() == [[0, 0], [1, 1], [2, 2]]

    with pytest.raises(StopIteration):
        next(batch_iter)


@pytest.mark.experimental
# noinspection PyTestParametrized
@pytest.mark.parametrize(
    "obs_range,var_range,X_value_gen,use_eager_fetch",
    [(6, 3, pytorch_x_value_gen, use_eager_fetch) for use_eager_fetch in (True, False)],
)
def test_batching__empty_query_result(soma_experiment: Experiment, use_eager_fetch: bool) -> None:
    exp_data_pipe = ExperimentDataPipe(
        soma_experiment,
        measurement_name="RNA",
        X_name="raw",
        obs_query=AxisQuery(coords=([],)),
        obs_column_names=["label"],
        batch_size=3,
        use_eager_fetch=use_eager_fetch,
    )
    batch_iter = iter(exp_data_pipe)

    with pytest.raises(StopIteration):
        next(batch_iter)


@pytest.mark.experimental
# noinspection PyTestParametrized
@pytest.mark.parametrize(
    "obs_range,var_range,X_value_gen,use_eager_fetch",
    [(6, 3, pytorch_x_value_gen, use_eager_fetch) for use_eager_fetch in (True, False)],
)
def test_sparse_output__non_batched(soma_experiment: Experiment, use_eager_fetch: bool) -> None:
    exp_data_pipe = ExperimentDataPipe(
        soma_experiment,
        measurement_name="RNA",
        X_name="raw",
        obs_column_names=["label"],
        return_sparse_X=True,
        use_eager_fetch=use_eager_fetch,
    )
    batch_iter = iter(exp_data_pipe)

    batch = next(batch_iter)
    assert isinstance(batch[1], Tensor)
    assert batch[0].to_dense().tolist() == [0, 1, 0]


@pytest.mark.experimental
# noinspection PyTestParametrized
@pytest.mark.parametrize(
    "obs_range,var_range,X_value_gen,use_eager_fetch",
    [(6, 3, pytorch_x_value_gen, use_eager_fetch) for use_eager_fetch in (True, False)],
)
def test_sparse_output__batched(soma_experiment: Experiment, use_eager_fetch: bool) -> None:
    exp_data_pipe = ExperimentDataPipe(
        soma_experiment,
        measurement_name="RNA",
        X_name="raw",
        obs_column_names=["label"],
        batch_size=3,
        return_sparse_X=True,
        use_eager_fetch=use_eager_fetch,
    )
    batch_iter = iter(exp_data_pipe)

    batch = next(batch_iter)
    assert isinstance(batch[1], Tensor)
    assert batch[0].to_dense().tolist() == [[0, 1, 0], [1, 0, 1], [0, 1, 0]]


@pytest.mark.experimental
# noinspection PyTestParametrized,DuplicatedCode
@pytest.mark.parametrize(
    "obs_range,var_range,X_value_gen,use_eager_fetch",
    [(10, 1, pytorch_x_value_gen, use_eager_fetch) for use_eager_fetch in (True, False)],
)
def test_batching__partial_soma_batches_are_concatenated(soma_experiment: Experiment, use_eager_fetch: bool) -> None:
    exp_data_pipe = ExperimentDataPipe(
        soma_experiment,
        measurement_name="RNA",
        X_name="raw",
        obs_column_names=[],
        batch_size=3,
        # set SOMA batch read size such that PyTorch batches will span the tail and head of two SOMA batches
        soma_chunk_size=4,
        use_eager_fetch=use_eager_fetch,
    )

    full_result = list(exp_data_pipe)

    assert [len(batch[1]) for batch in full_result] == [3, 3, 3, 1]


@pytest.mark.experimental
# noinspection PyTestParametrized
@pytest.mark.parametrize("obs_range,var_range,X_value_gen", [(3, 3, pytorch_x_value_gen)])
def test_encoders(soma_experiment: Experiment) -> None:
    exp_data_pipe = ExperimentDataPipe(
        soma_experiment,
        measurement_name="RNA",
        X_name="raw",
        obs_column_names=["label"],
        batch_size=3,
    )
    batch_iter = iter(exp_data_pipe)

    batch = next(batch_iter)
    assert isinstance(batch[1], Tensor)

    labels_encoded = batch[1][:, 1]
    labels_decoded = exp_data_pipe.obs_encoders["label"].inverse_transform(labels_encoded)
    assert labels_decoded.tolist() == ["0", "1", "2"]


@pytest.mark.experimental
@pytest.mark.skipif(
    (sys.version_info.major, sys.version_info.minor) == (3, 9), reason="fails intermittently with OOM error for 3.9"
)
# noinspection PyTestParametrized
@pytest.mark.parametrize("obs_range,var_range,X_value_gen", [(6, 3, pytorch_x_value_gen)])
def test_multiprocessing__returns_full_result(soma_experiment: Experiment) -> None:
    """Tests the ExperimentDataPipe provides all data, as collected from multiple processes that are managed by a
    PyTorch DataLoader with multiple workers configured."""

    dp = ExperimentDataPipe(
        soma_experiment,
        measurement_name="RNA",
        X_name="raw",
        obs_column_names=["label"],
        soma_chunk_size=3,  # two chunks, one per worker
    )
    # Note we're testing the ExperimentDataPipe via a DataLoader, since this is what sets up the multiprocessing
    dl = experiment_dataloader(dp, num_workers=2)

    full_result = list(iter(dl))

    soma_joinids = [t[1][0].item() for t in full_result]
    assert sorted(soma_joinids) == list(range(6))


@pytest.mark.experimental
# noinspection PyTestParametrized
@pytest.mark.parametrize("obs_range,var_range,X_value_gen", [(6, 3, pytorch_x_value_gen)])
def test_distributed__returns_data_partition_for_rank(soma_experiment: Experiment) -> None:
    """Tests pytorch._partition_obs_joinids() behavior in a simulated PyTorch distributed processing mode,
    using mocks to avoid having to do real PyTorch distributed setup."""

    with patch("cellxgene_census.experimental.ml.pytorch.dist.is_initialized") as mock_dist_is_initialized, patch(
        "cellxgene_census.experimental.ml.pytorch.dist.get_rank"
    ) as mock_dist_get_rank, patch(
        "cellxgene_census.experimental.ml.pytorch.dist.get_world_size"
    ) as mock_dist_get_world_size:
        mock_dist_is_initialized.return_value = True
        mock_dist_get_rank.return_value = 1
        mock_dist_get_world_size.return_value = 3

        dp = ExperimentDataPipe(
            soma_experiment, measurement_name="RNA", X_name="raw", obs_column_names=["label"], soma_chunk_size=2
        )
        full_result = list(iter(dp))

        soma_joinids = [t[1][0].item() for t in full_result]

        # Of the 6 obs rows, the PyTorch process of rank 1 should get [2, 3]
        # (rank 0 gets [0, 1], rank 2 gets [4, 5])
        assert sorted(soma_joinids) == [2, 3]


@pytest.mark.experimental
# noinspection PyTestParametrized
@pytest.mark.parametrize("obs_range,var_range,X_value_gen", [(12, 3, pytorch_x_value_gen)])
def test_distributed_and_multiprocessing__returns_data_partition_for_rank(soma_experiment: Experiment) -> None:
    """Tests pytorch._partition_obs_joinids() behavior in a simulated PyTorch distributed processing mode and
    DataLoader multiprocessing mode, using mocks to avoid having to do distributed pytorch
    setup or real DataLoader multiprocessing."""

    with patch("torch.utils.data.get_worker_info") as mock_get_worker_info, patch(
        "cellxgene_census.experimental.ml.pytorch.dist.is_initialized"
    ) as mock_dist_is_initialized, patch(
        "cellxgene_census.experimental.ml.pytorch.dist.get_rank"
    ) as mock_dist_get_rank, patch(
        "cellxgene_census.experimental.ml.pytorch.dist.get_world_size"
    ) as mock_dist_get_world_size:
        mock_get_worker_info.return_value = WorkerInfo(id=1, num_workers=2, seed=1234)
        mock_dist_is_initialized.return_value = True
        mock_dist_get_rank.return_value = 1
        mock_dist_get_world_size.return_value = 3

        dp = ExperimentDataPipe(
            soma_experiment, measurement_name="RNA", X_name="raw", obs_column_names=["label"], soma_chunk_size=2
        )

        full_result = list(iter(dp))

        soma_joinids = [t[1][0].item() for t in full_result]

        # Of the 12 obs rows, the PyTorch process of rank 1 should get [4..7], and then within that partition,
        # the 2nd DataLoader process should get the second half of the rank's partition, which is just [6, 7]
        # (rank 0 gets [0..3], rank 2 gets [8..11])
        assert sorted(soma_joinids) == [6, 7]


@pytest.mark.experimental
# noinspection PyTestParametrized,DuplicatedCode
@pytest.mark.parametrize(
    "obs_range,var_range,X_value_gen,use_eager_fetch",
    [(3, 3, pytorch_x_value_gen, use_eager_fetch) for use_eager_fetch in (True, False)],
)
def test_experiment_dataloader__non_batched(soma_experiment: Experiment, use_eager_fetch: bool) -> None:
    dp = ExperimentDataPipe(
        soma_experiment,
        measurement_name="RNA",
        X_name="raw",
        obs_column_names=["label"],
        use_eager_fetch=use_eager_fetch,
    )
    dl = experiment_dataloader(dp)
    torch_data = [row for row in dl]

    row = torch_data[0]
    assert row[0].to_dense().tolist() == [0, 1, 0]
    assert row[1].tolist() == [0, 0]


@pytest.mark.experimental
# noinspection PyTestParametrized,DuplicatedCode
@pytest.mark.parametrize(
    "obs_range,var_range,X_value_gen,use_eager_fetch",
    [(6, 3, pytorch_x_value_gen, use_eager_fetch) for use_eager_fetch in (True, False)],
)
def test_experiment_dataloader__batched(soma_experiment: Experiment, use_eager_fetch: bool) -> None:
    dp = ExperimentDataPipe(
        soma_experiment,
        measurement_name="RNA",
        X_name="raw",
        obs_column_names=["label"],
        batch_size=3,
        use_eager_fetch=use_eager_fetch,
    )
    dl = experiment_dataloader(dp)
    torch_data = [row for row in dl]

    batch = torch_data[0]
    assert batch[0].to_dense().tolist() == [[0, 1, 0], [1, 0, 1], [0, 1, 0]]
    assert batch[1].tolist() == [[0, 0], [1, 1], [2, 2]]


@pytest.mark.experimental
# noinspection PyTestParametrized,DuplicatedCode
@pytest.mark.parametrize(
    "obs_range,var_range,X_value_gen,use_eager_fetch",
    [(6, 3, pytorch_x_value_gen, use_eager_fetch) for use_eager_fetch in (True, False)],
)
def test__X_tensor_dtype_matches_X_matrix(soma_experiment: Experiment, use_eager_fetch: bool) -> None:
    dp = ExperimentDataPipe(
        soma_experiment,
        measurement_name="RNA",
        X_name="raw",
        obs_column_names=["label"],
        batch_size=3,
        use_eager_fetch=use_eager_fetch,
    )
    torch_data = next(iter(dp))

    assert torch_data[0].dtype == float32


@pytest.mark.experimental
# noinspection PyTestParametrized,DuplicatedCode
@pytest.mark.parametrize("obs_range,var_range,X_value_gen", [(10, 1, pytorch_x_value_gen)])
def test__pytorch_splitting(soma_experiment: Experiment) -> None:
    dp = ExperimentDataPipe(soma_experiment, measurement_name="RNA", X_name="raw", obs_column_names=["label"])
    dp_train, dp_test = dp.random_split(weights={"train": 0.7, "test": 0.3}, seed=1234)
    dl = experiment_dataloader(dp_train)

    all_rows = list(iter(dl))
    assert len(all_rows) == 7


@pytest.mark.experimental
# noinspection PyTestParametrized,DuplicatedCode
@pytest.mark.parametrize("obs_range,var_range,X_value_gen", [(16, 1, pytorch_seq_x_value_gen)])
def test__shuffle(soma_experiment: Experiment) -> None:
    dp = ExperimentDataPipe(
        soma_experiment, measurement_name="RNA", X_name="raw", obs_column_names=["label"], shuffle=True
    )

    all_rows = list(iter(dp))

    soma_joinids = [row[1][0].item() for row in all_rows]
    X_values = [row[0][0].item() for row in all_rows]

    # same elements
    assert set(soma_joinids) == set(range(16))
    # not ordered! (...with a `1/16!` probability of being ordered)
    assert soma_joinids != list(range(16))
    # randomizes X in same order as obs
    # note: X values were explicitly set to match obs_joinids to allow for this simple assertion
    assert X_values == soma_joinids


@pytest.mark.experimental
@pytest.mark.skip(reason="Not implemented")
def test_experiment_dataloader__multiprocess_sparse_matrix__fails() -> None:
    pass


@pytest.mark.experimental
@pytest.mark.skip(reason="Not implemented")
def test_experiment_dataloader__multiprocess_dense_matrix__ok() -> None:
    pass


@pytest.mark.experimental
@patch("cellxgene_census.experimental.ml.pytorch.ExperimentDataPipe")
def test_experiment_dataloader__unsupported_params__fails(dummy_exp_data_pipe: ExperimentDataPipe) -> None:
    with pytest.raises(ValueError):
        experiment_dataloader(dummy_exp_data_pipe, shuffle=True)
    with pytest.raises(ValueError):
        experiment_dataloader(dummy_exp_data_pipe, batch_size=3)
    with pytest.raises(ValueError):
        experiment_dataloader(dummy_exp_data_pipe, batch_sampler=[])
    with pytest.raises(ValueError):
        experiment_dataloader(dummy_exp_data_pipe, sampler=[])
    with pytest.raises(ValueError):
        experiment_dataloader(dummy_exp_data_pipe, collate_fn=lambda x: x)



import pickle
from typing import Any, Dict, Optional, Sequence, Set

import numpy as np
import numpy.typing as npt
import scipy
import tiledbsoma

from .cell_dataset_builder import CellDatasetBuilder


class GeneformerTokenizer(CellDatasetBuilder):
    """
    Generate a Hugging Face `Dataset` containing Geneformer token sequences for each
    cell in CELLxGENE Census ExperimentAxisQuery results (human).

    This class requires the Geneformer package to be installed separately with:
    `pip install git+https://huggingface.co/ctheodoris/Geneformer@8df5dc1`

    Example usage:

    ```
    import cellxgene_census
    import tiledbsoma
    from cellxgene_census.experimental.ml.huggingface import GeneformerTokenizer

    with cellxgene_census.open_soma(census_version="latest") as census:
        with GeneformerTokenizer(
            census["census_data"]["homo_sapiens"],
            # set obs_query to define some subset of Census cells:
            obs_query=tiledbsoma.AxisQuery(value_filter="is_primary_data == True and tissue_general == 'tongue'"),
            obs_column_names=(
                "soma_joinid",
                "cell_type_ontology_term_id",
            ),
        ) as tokenizer:
            dataset = tokenizer.build()
    ```

    Dataset item contents:
    - `input_ids`: Geneformer token sequence for the cell
    - `length`: Length of the token sequence
    - and the specified `obs_column_names` (cell metadata from the experiment obs dataframe)
    """

    obs_column_names: Set[str]
    max_input_tokens: int

    # set of gene soma_joinids corresponding to genes modeled by Geneformer:
    model_gene_ids: npt.NDArray[np.int64]
    model_gene_tokens: npt.NDArray[np.int64]  # token for each model_gene_id
    model_gene_medians: npt.NDArray[np.float64]  # float for each model_gene_id

    def __init__(
        self,
        experiment: tiledbsoma.Experiment,
        *,
        obs_column_names: Optional[Sequence[str]] = None,
        obs_attributes: Optional[Sequence[str]] = None,
        max_input_tokens: int = 2048,
        token_dictionary_file: str = "",
        gene_median_file: str = "",
        **kwargs: Any,
    ) -> None:
        """
        - `experiment`: Census Experiment to query
        - `obs_query`: obs AxisQuery defining the set of Census cells to process (default all)
        - `obs_column_names`: obs dataframe columns (cell metadata) to propagate into attributes
           of each Dataset item
        - `max_input_tokens`: maximum length of Geneformer input token sequence (default 2048)
        - `token_dictionary_file`, `gene_median_file`: pickle files supplying the mapping of
          Ensembl human gene IDs onto Geneformer token numbers and median expression values.
          By default, these will be loaded from the Geneformer package.
        """
        if obs_attributes:  # old name of obs_column_names
            obs_column_names = obs_attributes

        self.max_input_tokens = max_input_tokens
        self.obs_column_names = set(obs_column_names) if obs_column_names else set()
        self._load_geneformer_data(experiment, token_dictionary_file, gene_median_file)
        super().__init__(
            experiment,
            measurement_name="RNA",
            layer_name="raw",
            **kwargs,
        )

    def _load_geneformer_data(
        self, experiment: tiledbsoma.Experiment, token_dictionary_file: str, gene_median_file: str
    ) -> None:
        """
        Load (1) the experiment's genes dataframe and (2) Geneformer's static data
        files for gene tokens and median expression; then, intersect them to compute
        self.model_gene_{ids,tokens,medians}
        """
        # TODO: this work could be reused for all queries on this experiment

        genes_df = experiment.ms["RNA"].var.read(column_names=["soma_joinid", "feature_id"]).concat().to_pandas()

        if not (token_dictionary_file and gene_median_file):
            try:
                import geneformer
            except ImportError:
                # pyproject.toml can't express Geneformer git+https dependency
                raise ImportError(
                    "Please install Geneformer with: "
                    "pip install git+https://huggingface.co/ctheodoris/Geneformer@8df5dc1"
                ) from None
            if not token_dictionary_file:
                token_dictionary_file = geneformer.tokenizer.TOKEN_DICTIONARY_FILE
            if not gene_median_file:
                gene_median_file = geneformer.tokenizer.GENE_MEDIAN_FILE
        with open(token_dictionary_file, "rb") as f:
            gene_token_dict = pickle.load(f)
        with open(gene_median_file, "rb") as f:
            gene_median_dict = pickle.load(f)

        # compute model_gene_{ids,tokens,medians} by joining genes_df with Geneformer's
        # dicts
        model_gene_ids = []
        model_gene_tokens = []
        model_gene_medians = []
        for gene_id, row in genes_df.iterrows():
            ensg = row["feature_id"]  # ENSG... gene id, which keys Geneformer's dicts
            if ensg in gene_token_dict:
                model_gene_ids.append(gene_id)
                model_gene_tokens.append(gene_token_dict[ensg])
                model_gene_medians.append(gene_median_dict[ensg])
        self.model_gene_ids = np.array(model_gene_ids, dtype=np.int64)
        self.model_gene_tokens = np.array(model_gene_tokens, dtype=np.int64)
        self.model_gene_medians = np.array(model_gene_medians, dtype=np.float64)

        assert len(np.unique(self.model_gene_ids)) == len(self.model_gene_ids)
        assert len(np.unique(self.model_gene_tokens)) == len(self.model_gene_tokens)
        assert np.all(self.model_gene_medians > 0)
        # Geneformer models protein-coding and miRNA genes, so the intersection should
        # be somewhere a little north of 20K.
        assert len(self.model_gene_ids) > 20_000

        # Precompute a vector by which we'll multiply each cell's expression vector.
        # The denominator normalizes by Geneformer's median expression values.
        # The numerator 10K factor follows Geneformer's tokenizer; theoretically it doesn't affect
        # affect the rank order, but is probably intended to help with numerical precision.
        self.model_gene_medians_factor = 10_000.0 / self.model_gene_medians

    def __enter__(self) -> "GeneformerTokenizer":
        super().__enter__()
        # On context entry, load the necessary cell metadata (obs_df)
        obs_column_names = list(self.obs_column_names)
        if "soma_joinid" not in self.obs_column_names:
            obs_column_names.append("soma_joinid")
        self.obs_df = self.obs(column_names=obs_column_names).concat().to_pandas().set_index("soma_joinid")
        return self

    def cell_item(self, cell_joinid: int, cell_Xrow: scipy.sparse.csr_matrix) -> Dict[str, Any]:
        """
        Given the expression vector for one cell, compute the Dataset item providing
        the Geneformer inputs (token sequence and metadata).
        """
        # project cell_Xrow onto model_gene_ids and normalize by row sum.
        # notice we divide by the total count of the complete row (not only of the projected
        # values); this follows Geneformer's internal tokenizer.
        model_counts = cell_Xrow[:, self.model_gene_ids].multiply(1.0 / cell_Xrow.sum())
        assert isinstance(model_counts, scipy.sparse.csr_matrix), type(model_counts)
        # assert len(model_counts.data) == np.count_nonzero(model_counts.data)
        model_expr = model_counts.multiply(self.model_gene_medians_factor)
        assert isinstance(model_expr, scipy.sparse.coo_matrix), type(model_expr)
        # assert len(model_expr.data) == np.count_nonzero(model_expr.data)

        # figure the resulting tokens in descending order of model_expr
        # (use sparse model_expr.{col,data} to naturally exclude undetected genes)
        token_order = model_expr.col[np.argsort(-model_expr.data)[: self.max_input_tokens]]
        input_ids = self.model_gene_tokens[token_order]

        ans = {"input_ids": input_ids, "length": len(input_ids)}
        # add the requested obs attributes
        for attr in self.obs_column_names:
            if attr != "soma_joinid":
                ans[attr] = self.obs_df.at[cell_joinid, attr]
            else:
                ans["soma_joinid"] = cell_joinid
        return ans



from typing import List

import numpy as np
import pytest
import tiledbsoma as soma

import cellxgene_census


@pytest.fixture
def census() -> soma.Collection:
    return cellxgene_census.open_soma(census_version="latest")


@pytest.mark.live_corpus
def test_get_anndata_value_filter(census: soma.Collection) -> None:
    with census:
        ad = cellxgene_census.get_anndata(
            census,
            organism="Mus musculus",
            obs_value_filter="tissue_general == 'vasculature'",
            var_value_filter="feature_name in ['Gm53058', '0610010K14Rik']",
            column_names={
                "obs": ["soma_joinid", "cell_type", "tissue", "tissue_general", "assay"],
                "var": ["soma_joinid", "feature_id", "feature_name", "feature_length"],
            },
        )

    assert ad is not None
    assert ad.n_vars == 2
    assert ad.n_obs > 0
    assert (ad.obs.tissue_general == "vasculature").all()
    assert set(ad.var.feature_name) == {"Gm53058", "0610010K14Rik"}


@pytest.mark.live_corpus
def test_get_anndata_coords(census: soma.Collection) -> None:
    with census:
        ad = cellxgene_census.get_anndata(
            census, organism="Mus musculus", obs_coords=slice(1000), var_coords=slice(2000)
        )

    assert ad is not None
    assert ad.n_vars == 2001
    assert ad.n_obs == 1001


@pytest.mark.live_corpus
def test_get_anndata_allows_missing_obs_or_var_filter(census: soma.Collection) -> None:
    # This test is slightly sensitive to the live data, in that it assumes the
    # existance of certain cell tissue labels and gene feature ids.
    with census:
        mouse = census["census_data"]["mus_musculus"]

        adata = cellxgene_census.get_anndata(census, organism="Mus musculus", obs_value_filter="tissue == 'aorta'")
        assert adata.n_obs == len(
            mouse.obs.read(value_filter="tissue == 'aorta'", column_names=["soma_joinid"]).concat()
        )
        assert adata.n_vars == len(mouse.ms["RNA"].var.read(column_names=["soma_joinid"]).concat())

        adata = cellxgene_census.get_anndata(
            census,
            organism="Mus musculus",
            obs_coords=slice(10000),
            var_value_filter="feature_id == 'ENSMUSG00000069581'",
        )
        assert adata.n_obs == 10001
        assert adata.n_vars == 1


@pytest.mark.live_corpus
@pytest.mark.parametrize("layer", ["raw", "normalized"])
def test_get_anndata_x_layer(census: soma.Collection, layer: str) -> None:
    with census:
        ad = cellxgene_census.get_anndata(
            census,
            organism="Homo sapiens",
            X_name=layer,
            obs_coords=slice(100),
            var_coords=slice(200),
        )

    assert ad.X.shape == (101, 201)
    assert len(ad.layers) == 0


@pytest.mark.live_corpus
@pytest.mark.parametrize("layers", [["raw", "normalized"], ["normalized", "raw"]])
def test_get_anndata_two_layers(census: soma.Collection, layers: List[str]) -> None:
    with census:
        ad_primary_layer_in_X = cellxgene_census.get_anndata(
            census,
            organism="Homo sapiens",
            X_name=layers[0],
            obs_coords=slice(100),
            var_coords=slice(200),
        )

        ad_secondary_layer_in_X = cellxgene_census.get_anndata(
            census,
            organism="Homo sapiens",
            X_name=layers[1],
            obs_coords=slice(100),
            var_coords=slice(200),
        )

        ad_multiple_layers = cellxgene_census.get_anndata(
            census,
            organism="Homo sapiens",
            X_name=layers[0],
            X_layers=[layers[1]],
            obs_coords=slice(100),
            var_coords=slice(200),
        )

    assert layers[1] in ad_multiple_layers.layers
    assert ad_multiple_layers.X.shape == (101, 201)
    assert ad_multiple_layers.layers[layers[1]].shape == (101, 201)

    # Assert that matrices of multilayer anndata are equal to one-layer-at-time anndatas
    assert np.array_equal(ad_multiple_layers.X.data, ad_primary_layer_in_X.X.data)
    assert np.array_equal(ad_multiple_layers.layers[layers[1]].data, ad_secondary_layer_in_X.X.data)


@pytest.mark.live_corpus
def test_get_anndata_wrong_layer_names(census: soma.Collection) -> None:
    with census:
        with pytest.raises(ValueError) as raise_info:
            cellxgene_census.get_anndata(
                census,
                organism="Homo sapiens",
                X_name="this_layer_name_is_bad",
                obs_coords=slice(100),
                var_coords=slice(200),
            )

            assert raise_info.value.args[0] == "Unknown X layer name"

        with pytest.raises(ValueError) as raise_info:
            cellxgene_census.get_anndata(
                census,
                organism="Homo sapiens",
                X_name="raw",
                X_layers=["this_layer_name_is_bad"],
                obs_coords=slice(100),
                var_coords=slice(200),
            )

            assert raise_info.value.args[0] == "Unknown X layer name"


@pytest.mark.skip(reason="Enable when obsm is available in a live Census distribution.")
@pytest.mark.live_corpus
@pytest.mark.parametrize("obsm_layer", ["scvi", "geneformer"])
def test_get_anndata_obsm_one_layer(census: soma.Collection, obsm_layer: str) -> None:
    with census:
        ad = cellxgene_census.get_anndata(
            census,
            organism="Homo sapiens",
            X_name="raw",
            obs_coords=slice(100),
            var_coords=slice(200),
            obsm_layers=[obsm_layer],
        )

    assert len(ad.obsm.keys()) == 1
    assert obsm_layer in ad.obsm.keys()
    assert ad.obsm[obsm_layer].shape[0] == 100


@pytest.mark.skip(reason="Enable when obsm is available in a live Census distribution.")
@pytest.mark.live_corpus
@pytest.mark.parametrize("obsm_layers", [["scvi", "geneformer"]])
def test_get_anndata_obsm_two_layers(census: soma.Collection, obsm_layers: List[str]) -> None:
    with census:
        ad = cellxgene_census.get_anndata(
            census,
            organism="Homo sapiens",
            X_name="raw",
            obs_coords=slice(100),
            var_coords=slice(200),
            obsm_layers=obsm_layers,
        )

    assert len(ad.obsm.keys()) == 2
    for obsm_layer in obsm_layers:
        assert obsm_layer in ad.obsm.keys()
        assert ad.obsm[obsm_layer].shape[0] == 100



from typing import Any

import pytest
import requests_mock as rm
import s3fs

import cellxgene_census
from cellxgene_census._release_directory import CELL_CENSUS_MIRRORS_DIRECTORY_URL, CELL_CENSUS_RELEASE_DIRECTORY_URL

# This test fixture contains 3 releases: 1 "latest" and 2 "LTS". Of the "LTS" releases, one is aliased to "stable"
# and one is "retracted", and both are aliased with "V#" aliases. The ordering of the releases is
# explicitly set to verify that the directory is sorted correctly (i.e. they are not in the desired order here).
# There is also a "dangling" tag, to verify that we handle this case correctly.
#
# TODO: Break this into multiple fixtures to test different scenarios
DIRECTORY_JSON = {
    "2022-10-01": {
        "release_date": "2022-10-30",
        "release_build": "2022-10-01",
        "flags": {"lts": True},
        "soma": {
            "uri": "s3://cellxgene-data-public/cell-census/2022-10-01/soma/",
            "s3_region": "us-west-2",
        },
        "h5ads": {
            "uri": "s3://cellxgene-data-public/cell-census/2022-10-01/h5ads/",
            "s3_region": "us-west-2",
        },
    },
    "2022-09-01": {
        "release_date": "2022-09-30",
        "release_build": "2022-09-01",
        "flags": {"lts": True, "retracted": True},
        "do_not_delete": True,
        "retraction": {
            "date": "2022-11-15",
            "reason": "mistakes happen",
            "info_permalink": "http://cellxgene.com/census/apologies",
        },
        "soma": {
            "uri": "s3://cellxgene-data-public/cell-census/2022-09-01/soma/",
            "s3_region": "us-west-2",
        },
        "h5ads": {"uri": "s3://cellxgene-data-public/cell-census/2022-09-01/h5ads/", "s3_region": "us-west-2"},
    },
    # Ordered the latest release to be last, to verify it is explicitly sorted
    "2022-11-01": {
        "release_date": "2022-11-30",
        "release_build": "2022-11-01",
        "do_not_delete": True,
        "soma": {
            "uri": "s3://cellxgene-data-public/cell-census/2022-11-01/soma/",
            "s3_region": "us-west-2",
        },
        "h5ads": {
            "uri": "s3://cellxgene-data-public/cell-census/2022-11-01/h5ads/",
            "s3_region": "us-west-2",
        },
    },
    # An explicitly dangling tag, to confirm we handle correct
    # Underscore indicates expected failure to test below
    "_dangling": "no-such-tag",
    # Aliases placed at bottom, to verify these are explicitly sorted to the top
    "stable": "V2",
    "latest": "2022-11-01",
    "V2": "2022-10-01",
    "V1": "2022-09-01",
}

MIRRORS_JSON = {
    "default": "AWS-S3-us-west-2",
    "AWS-S3-us-west-2": {"provider": "S3", "base_uri": "s3://cellxgene-data-public/", "region": "us-west-2"},
}


@pytest.fixture
def directory_mock(requests_mock: rm.Mocker) -> Any:
    return requests_mock.get(CELL_CENSUS_RELEASE_DIRECTORY_URL, json=DIRECTORY_JSON)


@pytest.fixture
def mirrors_mock(requests_mock: rm.Mocker) -> Any:
    return requests_mock.get(CELL_CENSUS_MIRRORS_DIRECTORY_URL, json=MIRRORS_JSON)


def test_get_census_version_directory(directory_mock: Any) -> None:
    directory = cellxgene_census.get_census_version_directory()

    assert isinstance(directory, dict)
    assert len(directory) > 0
    assert all(isinstance(k, str) for k in directory.keys())
    assert all(isinstance(v, dict) for v in directory.values())

    assert "_dangling" not in directory

    assert directory["2022-11-01"] == DIRECTORY_JSON["2022-11-01"]
    assert directory["2022-10-01"] == DIRECTORY_JSON["2022-10-01"]
    assert "2022-09-01" not in directory  # retracted excluded by default

    assert directory["latest"] == DIRECTORY_JSON["2022-11-01"]
    assert directory["stable"] == DIRECTORY_JSON["2022-10-01"]
    assert directory["V2"] == DIRECTORY_JSON["2022-10-01"]
    assert "V1" not in directory  # retracted excluded by default

    for tag in directory:
        assert directory[tag] == cellxgene_census.get_census_version_description(tag)

    # Verify that the directory is sorted according to this criteria:
    # 1. Aliases first
    # 2. Non aliases after, in reverse order
    dir_list = list(directory)
    assert dir_list == ["stable", "latest", "V2", "2022-11-01", "2022-10-01"]


def test_get_census_version_directory__lts_only(directory_mock: Any) -> None:
    directory = cellxgene_census.get_census_version_directory(lts=True)

    assert directory.keys() == {"stable", "V2", "2022-10-01"}


def test_get_census_version_directory__exclude_lts(directory_mock: Any) -> None:
    directory = cellxgene_census.get_census_version_directory(lts=False)

    assert directory.keys() == {"latest", "2022-11-01"}


def test_get_census_version_directory__include_retracted(directory_mock: Any) -> None:
    directory = cellxgene_census.get_census_version_directory(retracted=None)

    assert "V1" in directory
    assert "2022-09-01" in directory


def test_get_census_version_directory__retraction_info(directory_mock: Any) -> None:
    directory = cellxgene_census.get_census_version_directory(retracted=True)

    assert directory["2022-09-01"]["retraction"] == {
        "date": "2022-11-15",
        "reason": "mistakes happen",
        "info_permalink": "http://cellxgene.com/census/apologies",
    }

    assert directory["V1"]["retraction"] == {
        "date": "2022-11-15",
        "reason": "mistakes happen",
        "info_permalink": "http://cellxgene.com/census/apologies",
    }


def test_get_census_version_description_errors() -> None:
    with pytest.raises(ValueError):
        cellxgene_census.get_census_version_description(census_version="no/such/version/exists")


def test_get_census_mirrors_directory(mirrors_mock: Any) -> None:
    directory = cellxgene_census.get_census_mirror_directory()
    assert "default" not in directory
    assert "AWS-S3-us-west-2" in directory
    assert directory["AWS-S3-us-west-2"] == MIRRORS_JSON["AWS-S3-us-west-2"]


@pytest.mark.live_corpus
def test_live_directory_contents() -> None:
    # Sanity check that all directory contents are usable. This uses the
    # live directory, so it _could_ start failing without a code change.
    # But given the purpose of this package, that seems like a reasonable
    # tradeoff, as the data directory should never be "corrupt" or there
    # is widespread impact on users.

    fs = s3fs.S3FileSystem(anon=True, cache_regions=True)

    directory = cellxgene_census.get_census_version_directory()
    assert "latest" in directory

    for version, version_description in directory.items():
        with cellxgene_census.open_soma(census_version=version) as census:
            assert census is not None

        assert fs.exists(version_description["soma"]["uri"])
        assert fs.exists(version_description["h5ads"]["uri"])



"""
Acceptance tests for the Census.

NOTE: those marked `expensive` are not run in the CI as they are, well, expensive...

Several of them will not run to completion except on VERY large hosts.

Intended use:  periodically do a manual run, including the expensive tests, on an
appropriately large host.

See README.md for historical data.
"""

from typing import Any, Dict, Iterator, Optional, Tuple

import pyarrow as pa
import pytest
import tiledbsoma as soma

import cellxgene_census
from cellxgene_census._open import DEFAULT_TILEDB_CONFIGURATION


def make_context(census_version: str, config: Optional[Dict[str, Any]] = None) -> soma.SOMATileDBContext:
    config = config or {}
    version = cellxgene_census.get_census_version_description(census_version)
    s3_region = version["soma"].get("s3_region", "us-west-2")
    config.update({"vfs.s3.region": s3_region})
    config.update({"vfs.s3.no_sign_request": "true"})
    return soma.options.SOMATileDBContext().replace(**{"tiledb_config": config})


@pytest.mark.live_corpus
@pytest.mark.parametrize("organism", ["homo_sapiens", "mus_musculus"])
def test_load_axes(organism: str) -> None:
    """Verify axes can be loaded into a Pandas DataFrame"""
    census = cellxgene_census.open_soma(census_version="latest")

    # use subset of columns for speed
    obs_df = (
        census["census_data"][organism]
        .obs.read(column_names=["soma_joinid", "cell_type", "tissue"])
        .concat()
        .to_pandas()
    )
    assert len(obs_df)
    del obs_df

    var_df = census["census_data"][organism].ms["RNA"].var.read().concat().to_pandas()
    assert len(var_df)
    del var_df


def table_iter_is_ok(tbl_iter: Iterator[pa.Table], stop_after: Optional[int] = 2) -> bool:
    """
    Utility that verifies that the value is an iterator of pa.Table.

    Will only call __next__ as many times as the `stop_after` param specifies,
    or will read until end of iteration of it is None.
    """
    assert isinstance(tbl_iter, Iterator)
    for n, tbl in enumerate(tbl_iter):
        # keep things speedy by quitting early if stop_after specified
        if stop_after is not None and n > stop_after:
            break
        assert isinstance(tbl, pa.Table)
        assert len(tbl)

    return True


@pytest.mark.live_corpus
@pytest.mark.parametrize("organism", ["homo_sapiens", "mus_musculus"])
@pytest.mark.parametrize(
    ("stop_after", "ctx_config"),
    [
        pytest.param(2, None),
        pytest.param(None, DEFAULT_TILEDB_CONFIGURATION, marks=pytest.mark.expensive),
    ],
)
def test_incremental_read_obs(organism: str, stop_after: Optional[int], ctx_config: Optional[Dict[str, Any]]) -> None:
    """Verify that obs, var and X[raw] can be read incrementally, i.e., in chunks"""

    # ctx_config=None open census with a small (default) TileDB buffer size, which reduces
    # memory use, and makes it feasible to run in a GHA.
    ctx_config = ctx_config or {}
    context = make_context("latest", ctx_config)
    with cellxgene_census.open_soma(census_version="latest", context=context) as census:
        assert table_iter_is_ok(
            census["census_data"][organism].obs.read(column_names=["soma_joinid", "tissue"]),
            stop_after=stop_after,
        )


@pytest.mark.live_corpus
@pytest.mark.parametrize("organism", ["homo_sapiens", "mus_musculus"])
@pytest.mark.parametrize(
    ("stop_after", "ctx_config"),
    [
        pytest.param(2, None),
        pytest.param(None, DEFAULT_TILEDB_CONFIGURATION, marks=pytest.mark.expensive),
    ],
)
def test_incremental_read_var(organism: str, stop_after: Optional[int], ctx_config: Optional[Dict[str, Any]]) -> None:
    """Verify that var can be read incrementally, i.e., in chunks"""

    # ctx_config=None open census with a small (default) TileDB buffer size, which reduces
    # memory use, and makes it feasible to run in a GHA.
    ctx_config = ctx_config or {}
    context = make_context("latest", ctx_config)
    with cellxgene_census.open_soma(census_version="latest", context=context) as census:
        assert table_iter_is_ok(
            census["census_data"][organism].ms["RNA"].var.read(column_names=["soma_joinid", "feature_id"]),
            stop_after=stop_after,
        )


@pytest.mark.live_corpus
@pytest.mark.parametrize("organism", ["homo_sapiens", "mus_musculus"])
@pytest.mark.parametrize(
    ("stop_after", "ctx_config", "coords"),
    [
        pytest.param(2, None, (None, None)),
        pytest.param(
            None,
            DEFAULT_TILEDB_CONFIGURATION,
            (slice(0, 500_000), slice(None)),
            marks=pytest.mark.expensive,
        ),
        pytest.param(
            None,
            {"soma.init_buffer_bytes": 4 * 1024**3},
            (slice(0, 500_000), slice(None)),
            marks=pytest.mark.expensive,
        ),
        pytest.param(
            None,
            {"soma.init_buffer_bytes": 4 * 1024**3},
            (slice(0, 500_000), slice(0, 1_000)),
            marks=pytest.mark.expensive,
        ),
    ],
)
def test_incremental_read_X(
    organism: str,
    stop_after: Optional[int],
    ctx_config: Optional[Dict[str, Any]],
    coords: Optional[Tuple[slice, slice]],
) -> None:
    """Verify that obs, var and X[raw] can be read incrementally, i.e., in chunks"""

    ctx_config = ctx_config or {}
    context = make_context("latest", ctx_config)
    with cellxgene_census.open_soma(census_version="latest", context=context) as census:
        assert table_iter_is_ok(
            census["census_data"][organism].ms["RNA"].X["raw"].read(coords=coords).tables(),
            stop_after=stop_after,
        )


@pytest.mark.live_corpus
@pytest.mark.parametrize("organism", ["homo_sapiens", "mus_musculus"])
@pytest.mark.parametrize(
    "obs_value_filter",
    ["tissue=='aorta'", pytest.param("tissue=='brain'", marks=pytest.mark.expensive)],
)
@pytest.mark.parametrize("stop_after", [2, pytest.param(None, marks=pytest.mark.expensive)])
def test_incremental_query(organism: str, obs_value_filter: str, stop_after: Optional[int]) -> None:
    """Verify incremental read of query result."""
    # use default TileDB configuration
    with cellxgene_census.open_soma(census_version="latest") as census:
        with census["census_data"][organism].axis_query(
            measurement_name="RNA",
            obs_query=soma.AxisQuery(value_filter=obs_value_filter),
        ) as query:
            assert table_iter_is_ok(query.obs(), stop_after=stop_after)
            assert table_iter_is_ok(query.var(), stop_after=stop_after)
            assert table_iter_is_ok(query.X("raw").tables(), stop_after=stop_after)


@pytest.mark.live_corpus
@pytest.mark.parametrize("organism", ["homo_sapiens", "mus_musculus"])
@pytest.mark.parametrize(
    ("obs_value_filter", "obs_coords", "ctx_config"),
    [
        # small query, should be runable in CI
        pytest.param("tissue=='aorta'", None, DEFAULT_TILEDB_CONFIGURATION),
        # 10K cells, also small enough to run in CI
        pytest.param(None, slice(0, 10_000), DEFAULT_TILEDB_CONFIGURATION, id="First 10K cells"),
        # 100K cells, standard buffer size
        pytest.param(
            None,
            slice(0, 100_000),
            DEFAULT_TILEDB_CONFIGURATION,
            marks=pytest.mark.expensive,
            id="First 100K cells",
        ),
        # 250K cells, standard buffer size
        pytest.param(
            None,
            slice(0, 250_000),
            DEFAULT_TILEDB_CONFIGURATION,
            marks=pytest.mark.expensive,
            id="First 250K cells",
        ),
        # 500K cells, standard buffer size
        pytest.param(
            None,
            slice(0, 500_000),
            DEFAULT_TILEDB_CONFIGURATION,
            marks=pytest.mark.expensive,
            id="First 500K cells",
        ),
        # 750K cells, standard buffer size
        pytest.param(
            None,
            slice(0, 750_000),
            DEFAULT_TILEDB_CONFIGURATION,
            marks=pytest.mark.expensive,
            id="First 750K cells",
        ),
        # 1M cells, standard buffer size
        pytest.param(
            None,
            slice(0, 1_000_000),
            DEFAULT_TILEDB_CONFIGURATION,
            marks=pytest.mark.expensive,
            id="First 1M cells",
        ),
        # very common tissue, with standard buffer size
        pytest.param(
            "tissue_general=='brain'",
            slice(0, 2_000_000),
            DEFAULT_TILEDB_CONFIGURATION,
            marks=pytest.mark.expensive,
        ),
        # very common cell type, with standard buffer size
        pytest.param(
            "cell_type=='neuron'",
            slice(0, 2_000_000),
            DEFAULT_TILEDB_CONFIGURATION,
            marks=pytest.mark.expensive,
        ),
        # many primary cells, with big buffer size
        pytest.param(
            "is_primary_data==True",
            slice(0, 2_000_000),
            {"soma.init_buffer_bytes": 4 * 1024**3},
            marks=pytest.mark.expensive,
        ),
        #
        # a large enchilada, if not the whole thing, with a big buffer size
        pytest.param(
            None,
            slice(0, 2_000_000),
            {"soma.init_buffer_bytes": 4 * 1024**3},
            marks=pytest.mark.expensive,
        ),
    ],
)
def test_get_anndata(
    organism: str,
    obs_value_filter: Optional[str],
    obs_coords: Optional[slice],
    ctx_config: Optional[Dict[str, Any]],
) -> None:
    """Verify query and read into AnnData"""
    ctx_config = ctx_config or {}
    context = make_context("latest", ctx_config)
    with cellxgene_census.open_soma(census_version="latest", context=context) as census:
        ad = cellxgene_census.get_anndata(census, organism, obs_value_filter=obs_value_filter, obs_coords=obs_coords)
        assert ad is not None

        # sanity checks
        with census["census_data"][organism].axis_query(
            measurement_name="RNA",
            obs_query=soma.AxisQuery(value_filter=obs_value_filter, coords=(obs_coords,)),
        ) as query:
            assert ad.n_obs == query.n_obs
            assert ad.n_vars == query.n_vars



"""
API to facilitate preprocessing of SOMA datasets.
"""

from ._highly_variable_genes import get_highly_variable_genes, highly_variable_genes
from ._stats import mean_variance

__all__ = [
    "get_highly_variable_genes",
    "highly_variable_genes",
    "mean_variance",
]



# Copyright (c) 2022-2023 Chan Zuckerberg Initiative Foundation
#
# Licensed under the MIT License.

"""Versioning of Census builds

Methods to retrieve information about versions of the publicly hosted Census object.
"""
import typing
from collections import OrderedDict
from typing import Dict, Literal, Optional, Union, cast

import requests
from typing_extensions import NotRequired, TypedDict

"""
The following types describe the expected directory of Census builds, used
to bootstrap all data location requests.
"""
CensusVersionName = str  # census version name, e.g., "release-99", "2022-10-01-test", etc.
CensusLocator = TypedDict(
    "CensusLocator",
    {
        "uri": str,  # [deprecated: only used in census < 1.6.0] absolute resource URI.
        "relative_uri": str,  # resource URI (relative)
        "s3_region": Optional[str],  # [deprecated: only used in census < 1.6.0] if an S3 URI, has optional region
    },
)
CensusVersionRetraction = TypedDict(
    "CensusVersionRetraction",
    {
        "date": str,  # the date of retraction
        "reason": Optional[str],  # the reason for retraction
        "info_url": Optional[str],  # a permalink to more information
        "replaced_by": Optional[str],  # the census version that replaces this one
    },
)
ReleaseFlag = Literal["lts", "retracted"]
ReleaseFlags = Dict[ReleaseFlag, bool]
CensusVersionDescription = TypedDict(
    "CensusVersionDescription",
    {
        "release_date": Optional[str],  # date of release (deprecated)
        "release_build": str,  # date of build
        "soma": CensusLocator,  # SOMA objects locator
        "h5ads": CensusLocator,  # source H5ADs locator
        "flags": NotRequired[ReleaseFlags],  # flags for the release
        "retraction": NotRequired[CensusVersionRetraction],  # if retracted, details of the retraction
    },
)
CensusDirectory = Dict[CensusVersionName, Union[CensusVersionName, CensusVersionDescription]]

"""
A provider identifies a storage medium for the Census, which can either be a cloud provider or a local file.
A value of "unknown" can be specified if the provider isn't specified - the API will try to determine
the correct configuration based on the URI.
"""
Provider = Literal["S3", "file", "unknown"]

"""
A mirror identifies a location that can host the census artifacts. A dict of available mirrors exists
in the mirrors.json file, and looks like this:

{
    "default": "default-mirror",
    "default-mirror": {
        "provider": "S3",
        "base_uri": "s3://a-public-bucket/",
        "region": "us-west-2"
    }
}

"""
CensusMirrorName = str  # name of the mirror
CensusMirror = TypedDict(
    "CensusMirror",
    {
        "provider": Provider,  # provider of the mirror.
        "base_uri": str,  # base URI for the mirror location, e.g. s3://cellxgene-data-public/
        "region": Optional[str],  # region of the bucket or resource
    },
)

CensusMirrors = Dict[CensusMirrorName, Union[CensusMirrorName, CensusMirror]]

"""
A `ResolvedCensusLocator` represent an absolute location of a Census resource, including the provider info.
It is obtained by resolving a relative location against a specified mirror.
"""
ResolvedCensusLocator = TypedDict(
    "ResolvedCensusLocator",
    {
        "uri": str,  # resource URI (absolute)
        "region": Optional[str],  # if an S3 URI, has optional region
        "provider": str,  # Provider
    },
)


# URL for the default top-level directory of all public data
CELL_CENSUS_RELEASE_DIRECTORY_URL = "https://census.cellxgene.cziscience.com/cellxgene-census/v1/release.json"
CELL_CENSUS_MIRRORS_DIRECTORY_URL = "https://census.cellxgene.cziscience.com/cellxgene-census/v1/mirrors.json"


def get_census_version_description(census_version: str) -> CensusVersionDescription:
    """Get release description for given Census version, from the Census release directory.

    Args:
        census_version:
            The census version name.

    Returns:
        ``CensusVersionDescription`` - a dictionary containing a description of the release.

    Raises:
        ValueError: if unknown census_version value.

    Lifecycle:
        maturing

    See Also:
        :func:`get_census_version_directory`: returns the entire directory as a dict.

    Examples:
        >>> cellxgene_census.get_census_version_description("latest")
        {'release_date': None,
        'release_build': '2022-12-01',
        'soma': {'uri': 's3://cellxgene-data-public/cell-census/2022-12-01/soma/',
        's3_region': 'us-west-2'},
        'h5ads': {'uri': 's3://cellxgene-data-public/cell-census/2022-12-01/h5ads/',
        's3_region': 'us-west-2'}}
    """
    census_directory = get_census_version_directory()
    description = census_directory.get(census_version, None)
    if description is None:
        raise ValueError(f"Unable to locate Census version: {census_version}.")
    return description


def get_census_version_directory(
    *, lts: Optional[bool] = None, retracted: Optional[bool] = False
) -> Dict[CensusVersionName, CensusVersionDescription]:
    """
    Get the directory of Census versions currently available, optionally filtering by specified flags. If a filtering
    flag is not specified, Census versions will not be filtered by that flag. Defaults to including both "long-term
    stable" (LTS) and weekly Census versions, and excluding retracted versions.

    Params:
        lts: A filtering flag to either include or exclude long-term stable releases in the result. If None, no
         filtering is performed based on this flag. Defaults to None, which includes both LTS and non-LTS (weekly)
         versions.
        retracted: A filtering flag to either include or exclude retracted releases in the result. If None, no
         filtering is performed based on this flag. Defaults to False, which excludes retracted releases in the result.

    Returns:
        A dictionary that contains Census version names and their corresponding descriptions. Census versions are
        always named by their release date (``YYYY-MM-DD``) but may also have aliases. If an alias is specified,
        the Census version will appear multiple times in the dictionary, once under it's release date name,
        and again for each alias. Aliases may be: "stable", "latest", or "V#". The "stable" alias is used for the
        most recent LTS release, the "latest" alias is used for the most recent weekly release, and the "V#" aliases
        are used to identify LTS releases by a sequentially incrementing version number.

    Lifecycle:
        maturing

    See Also:
        :func:`get_census_version_description`: get description by census_version_name.

    Examples:
        Get all LTS and weekly versions, but exclude retracted LTS versions:

        >>> cellxgene_census.get_census_version_directory()
            {
                'stable': {
                    'release_date': None,
                    'release_build': '2022-11-29',
                    'soma': {'uri': 's3://cellxgene-data-public/cell-census/2022-11-29/soma/',
                             's3_region': 'us-west-2'},
                    'h5ads': {'uri': 's3://cellxgene-data-public/cell-census/2022-11-29/h5ads/',
                              's3_region': 'us-west-2'},
                    'flags': {'lts': True, 'retracted': False}
                },
                'latest': {
                    'release_date': None,
                    'release_build': '2022-12-01',
                    'soma': {'uri': 's3://cellxgene-data-public/cell-census/2022-12-01/soma/',
                             's3_region': 'us-west-2'},
                    'h5ads': {'uri': 's3://cellxgene-data-public/cell-census/2022-12-01/h5ads/',
                              's3_region': 'us-west-2'},
                    'flags': {'lts': True, 'retracted': False}
                },
                'V2': {
                    'release_date': None,
                    'release_build': '2022-11-29',
                    'soma': {'uri': 's3://cellxgene-data-public/cell-census/2022-11-29/soma/',
                             's3_region': 'us-west-2'},
                    'h5ads': {'uri': 's3://cellxgene-data-public/cell-census/2022-11-29/h5ads/',
                              's3_region': 'us-west-2'},
                    'flags': {'lts': True, 'retracted': False}
                },
                '2022-12-01': {
                    'release_date': None,
                    'release_build': '2022-12-01',
                    'soma': {'uri': 's3://cellxgene-data-public/cell-census/2022-12-01/soma/',
                             's3_region': 'us-west-2'},
                    'h5ads': {'uri': 's3://cellxgene-data-public/cell-census/2022-12-01/h5ads/',
                              's3_region': 'us-west-2'},
                    'flags': {'lts': False, 'retracted': False}
                },
                '2022-11-29': {
                    'release_date': None,
                    'release_build': '2022-11-29',
                    'soma': {'uri': 's3://cellxgene-data-public/cell-census/2022-11-29/soma/',
                             's3_region': 'us-west-2'},
                    'h5ads': {'uri': 's3://cellxgene-data-public/cell-census/2022-11-29/h5ads/',
                              's3_region': 'us-west-2'},
                    'flags': {'lts': True, 'retracted': False}
                }
            }

        Get only LTS versions that are not retracted:

        >>> cellxgene_census.get_census_version_directory(lts=True)
            {
                'stable': {
                    'release_date': None,
                    'release_build': '2022-11-29',
                    'soma': {'uri': 's3://cellxgene-data-public/cell-census/2022-11-29/soma/',
                             's3_region': 'us-west-2'},
                    'h5ads': {'uri': 's3://cellxgene-data-public/cell-census/2022-11-29/h5ads/',
                              's3_region': 'us-west-2'},
                    'flags': {'lts': True, 'retracted': False}
                },
                'V2': {
                    'release_date': None,
                    'release_build': '2022-11-29',
                    'soma': {'uri': 's3://cellxgene-data-public/cell-census/2022-11-29/soma/',
                             's3_region': 'us-west-2'},
                    'h5ads': {'uri': 's3://cellxgene-data-public/cell-census/2022-11-29/h5ads/',
                              's3_region': 'us-west-2'},
                    'flags': {'lts': True, 'retracted': False}
                },
                '2022-11-29': {
                    'release_date': None,
                    'release_build': '2022-11-29',
                    'soma': {'uri': 's3://cellxgene-data-public/cell-census/2022-11-29/soma/',
                             's3_region': 'us-west-2'},
                    'h5ads': {'uri': 's3://cellxgene-data-public/cell-census/2022-11-29/h5ads/',
                              's3_region': 'us-west-2'},
                    'flags': {'lts': True, 'retracted': False}
                }
            }

        Get only retracted releases:

        >>> cellxgene_census.get_census_version_directory(retracted=True)
            {
                'V1': {
                    'release_date': None,
                    'release_build': '2022-10-15',
                    'soma': {'uri': 's3://cellxgene-data-public/cell-census/2022-10-15/soma/',
                             's3_region': 'us-west-2'},
                    'h5ads': {'uri': 's3://cellxgene-data-public/cell-census/2022-10-15/h5ads/',
                              's3_region': 'us-west-2'},
                    'flags': {'lts': True, 'retracted': True},
                    'retraction': {
                        'date': '2022-10-30',
                        'reason': 'mistakes happen',
                        'info_url': 'http://cellxgene.com/census/errata/v1',
                        'replaced_by': 'V2'
                    },
                },
                '2022-10-15': {
                    'release_date': None,
                    'release_build': '2022-10-15',
                    'soma': {'uri': 's3://cellxgene-data-public/cell-census/2022-10-15/soma/',
                             's3_region': 'us-west-2'},
                    'h5ads': {'uri': 's3://cellxgene-data-public/cell-census/2022-10-15/h5ads/',
                              's3_region': 'us-west-2'},
                    'flags': {'lts': True, 'retracted': True},
                    'retraction': {
                        'date': '2022-10-30',
                        'reason': 'mistakes happen',
                        'info_url': 'http://cellxgene.com/census/errata/v1',
                        'replaced_by': 'V2'
                    }
                }
            }
    """
    response = requests.get(CELL_CENSUS_RELEASE_DIRECTORY_URL)
    response.raise_for_status()

    directory: CensusDirectory = cast(CensusDirectory, response.json())
    directory_out: CensusDirectory = {}
    aliases: typing.Set[CensusVersionName] = set()

    # Resolve all aliases for easier use
    for census_version_name in list(directory.keys()):
        # Strings are aliases for other census_version_name
        directory_value = directory[census_version_name]
        alias = None
        while isinstance(directory_value, str):
            alias = directory_value
            # resolve aliases
            if alias not in directory:
                # oops, dangling pointer -- drop original census_version_name
                directory.pop(census_version_name)
                break

            directory_value = directory[alias]

        if alias:
            aliases.add(census_version_name)

        # exclude aliases
        if not isinstance(directory_value, dict):
            continue

        # filter by release flags
        census_version_description = cast(CensusVersionDescription, directory_value)
        release_flags = cast(ReleaseFlags, {"lts": lts, "retracted": retracted})
        admitted = all(
            [
                census_version_description.get("flags", {}).get(flag_name, False) == release_flags[flag_name]
                for flag_name, flag_value in release_flags.items()
                if flag_value is not None
            ]
        )
        if not admitted:
            continue

        directory_out[census_version_name] = census_version_description.copy()

    # Cast is safe, as we have removed all aliases
    unordered_directory = cast(Dict[CensusVersionName, CensusVersionDescription], directory_out)

    # Sort by aliases and release date, descending
    aliased_releases = [(k, v) for k, v in unordered_directory.items() if k in aliases]
    concrete_releases = [(k, v) for k, v in unordered_directory.items() if k not in aliases]
    ordered_directory = OrderedDict()
    # Note: reverse sorting of aliases serendipitously orders the names we happen to use in a desirable manner:
    # "stable", "latest", "V#"). This will require a more explicit ordering if we change alias naming conventions.
    for k, v in sorted(aliased_releases, key=lambda k: k[0], reverse=True) + sorted(
        concrete_releases, key=lambda k: k[0], reverse=True
    ):
        ordered_directory[k] = v

    return ordered_directory


def get_census_mirror_directory() -> Dict[CensusMirrorName, CensusMirror]:
    """
    Get the directory of Census mirrors currently available.

    Returns:
        A dictionary that contains mirror names and their corresponding info,
        like the provider and the region.

    Lifecycle:
        maturing
    """
    mirrors = _get_census_mirrors()
    del mirrors["default"]
    return cast(Dict[CensusMirrorName, CensusMirror], mirrors)


def _get_census_mirrors() -> CensusMirrors:
    response = requests.get(CELL_CENSUS_MIRRORS_DIRECTORY_URL)
    response.raise_for_status()
    return cast(CensusMirrors, response.json())



# Copyright (c) 2022-2023 Chan Zuckerberg Initiative Foundation
#
# Licensed under the MIT License.

"""Open census and related datasets

Contains methods to open publicly hosted versions of Census object and access its source datasets.
"""
import logging
import os.path
import urllib.parse
from typing import Any, Dict, Optional, get_args

import s3fs
import tiledbsoma as soma

from ._release_directory import (
    CensusLocator,
    CensusMirror,
    Provider,
    ResolvedCensusLocator,
    _get_census_mirrors,
    get_census_version_description,
)
from ._util import _uri_join

DEFAULT_CENSUS_VERSION = "stable"

DEFAULT_TILEDB_CONFIGURATION: Dict[str, Any] = {
    # https://docs.tiledb.com/main/how-to/configuration#configuration-parameters
    "py.init_buffer_bytes": 1 * 1024**3,
    "soma.init_buffer_bytes": 1 * 1024**3,
    # S3 requests should not be signed, since we want to allow anonymous access
    "vfs.s3.no_sign_request": "true",
    "vfs.s3.region": "us-west-2",
}

api_logger = logging.getLogger("cellxgene_census")
api_logger.setLevel(logging.INFO)
api_logger.addHandler(logging.StreamHandler())


def _assert_mirror_supported(mirror: CensusMirror) -> None:
    """
    Verifies if the mirror is supported by this version of the census API.
    This method provides a proper error message in case an old version of the census
    tries to connect to an unsupported mirror.
    """
    if mirror["provider"] not in get_args(Provider):
        raise ValueError(
            f"Unsupported mirror provider: {mirror['provider']}. Try upgrading the cellxgene-census package to the latest version."
        )


def _resolve_census_locator(locator: CensusLocator, mirror: CensusMirror) -> ResolvedCensusLocator:
    _assert_mirror_supported(mirror)

    if locator.get("relative_uri"):
        uri = _uri_join(mirror["base_uri"], locator["relative_uri"])
        region = mirror["region"]
    else:
        uri = locator["uri"]
        region = locator.get("s3_region")
    return ResolvedCensusLocator(uri=uri, region=region, provider=mirror["provider"])


def _open_soma(
    locator: ResolvedCensusLocator, context: Optional[soma.options.SOMATileDBContext] = None
) -> soma.Collection:
    """
    Private. Merge config defaults and return open census as a soma Collection/context.
    """

    # if no user-defined context, cellxgene_census defaults take precedence over SOMA defaults
    context = context or get_default_soma_context()

    # A locator's S3 bucket and region are intrinsically coupled, so ensure that the context always uses the
    # locator's region, even if the passed-in context specifies an alternate S3 region
    if locator["provider"] == "S3":
        context = context.replace(tiledb_config={"vfs.s3.region": locator.get("region")})

    return soma.open(locator["uri"], mode="r", soma_type=soma.Collection, context=context)


def get_default_soma_context(tiledb_config: Optional[Dict[str, Any]] = None) -> soma.options.SOMATileDBContext:
    """Return a ``SOMATileDBContext`` with sensible defaults that can be further customized by the user.  The customized
    context can then be passed to ``open_soma(context=...)`` or a ``SOMAObject.open(context=...)`` method,
    such as ``tiledbsoma.Experiment.open()``.  Use the ``replace()`` method on the returned object to customize its
    settings further.

    Args:
        tiledb_config:
            A dictionary of TileDB configuration parameters. If specified, the parameters will override the
            defaults. If not specified, the default configuration will be returned.

    Returns:
        A ``SOMATileDBContext`` object with sensible defaults.

    Examples:

        To reduce the amount of memory used by TileDB-SOMA I/O operations:

        .. highlight:: python
        .. code-block:: python

            ctx = cellxgene_census.get_default_soma_context(
                tiledb_config={"py.init_buffer_bytes": 128 * 1024**2,
                               "soma.init_buffer_bytes": 128 * 1024**2})
            c = census.open_soma(uri="s3://my-private-bucket/census/soma", context=ctx)

        To access a copy of the Census located in a private bucket that is located in a different S3 region, use:

        .. highlight:: python
        .. code-block:: python

            ctx = cellxgene_census.get_default_soma_context(
                tiledb_config={"vfs.s3.no_sign_request": "false",
                               "vfs.s3.region": "us-east-1"})
            c = census.open_soma(uri="s3://my-private-bucket/census/soma", context=ctx)

    Lifecycle:
        experimental
    """

    tiledb_config = dict(DEFAULT_TILEDB_CONFIGURATION, **(tiledb_config or {}))
    return soma.options.SOMATileDBContext().replace(tiledb_config=tiledb_config)


def open_soma(
    *,
    census_version: Optional[str] = DEFAULT_CENSUS_VERSION,
    mirror: Optional[str] = None,
    uri: Optional[str] = None,
    tiledb_config: Optional[Dict[str, Any]] = None,
    context: Optional[soma.options.SOMATileDBContext] = None,
) -> soma.Collection:
    """Open the Census by version or URI.

    Args:
        census_version:
            The version of the Census, e.g. "latest" or "stable". Defaults to "stable".
        mirror:
            The mirror used to retrieve the Census. If not specified, a suitable mirror
            will be chosen automatically.
        uri:
            The URI containing the Census SOMA objects. If specified, will take precedence
            over ``census_version`` parameter.
        tiledb_config:
            A dictionary of TileDB configuration parameters that will be used to open the SOMA object.  Optional,
            defaults to None.  If specified, the parameters will override the default settings specified by
           ``get_default_soma_context().tiledb_config``.  Only one of the ``tiledb_config and ``context`` params
            can be specified.
        context:
            A custom :class:`SOMATileDBContext` that will be used to open the SOMA object.
            Optional, defaults to None.  Only one of the ``tiledb_config and ``context`` params can be specified.

    Returns:
        A SOMA Collection object containing the top-level census.
        It can be used as a context manager, which will automatically close upon exit.

    Raises:
        ValueError: if the census cannot be found, the URI cannot be opened, neither a URI
            or a version are specified, or an invalid mirror is provided.

    See Also:
        - :func:`get_source_h5ad_uri`: Look up the location of the source H5AD.
        - :func:`get_default_soma_context`: Get a default SOMA context that can be updated with custom settings and used
          as the ``context`` argument.

    Lifecycle:
        maturing

    Examples:
        Open the default Census version, using a context manager which will automatically
        close the Census upon exit of the context.

        >>> with cellxgene_census.open_soma() as census:
                ...

        Open and close:

        >>> census = cellxgene_census.open_soma()
            ...
            census.close()

        Open a specific Census by version:

        >>> with cellxgene_census.open_soma("2022-12-31") as census:
                ...

        Open a Census by S3 URI, rather than by version.

        >>> with cellxgene_census.open_soma(uri="s3://bucket/path") as census:
                ...

        Open a Census by path (file:// URI), rather than by version.

        >>> with cellxgene_census.open_soma(uri="/tmp/census") as census:
                ...

        Open a Census using a mirror.

        >>> with cellxgene_census.open_soma(mirror="s3-us-west-2") as census:
                ...

        Open a Census with a custom TileDB configuration setting.

        >>> with cellxgene_census.open_soma(tiledb_config={"py.init_buffer_bytes": 128 * 1024**2}) as census:
                ...
    """

    if tiledb_config is not None and context is not None:
        raise ValueError("Only one of tiledb_config and context can be specified.")

    if tiledb_config is not None:
        context = get_default_soma_context(tiledb_config=tiledb_config)

    if uri is not None:
        return _open_soma({"uri": uri, "region": None, "provider": "unknown"}, context)

    if census_version is None:
        raise ValueError("Must specify either a census version or an explicit URI.")

    mirrors = _get_census_mirrors()
    selected_mirror: CensusMirror
    if mirror is not None:
        if mirror not in mirrors:
            raise ValueError("Mirror not found.")
        selected_mirror = mirrors[mirror]  # type: ignore
    else:
        selected_mirror = mirrors[mirrors["default"]]  # type: ignore

    # TODO: Consider raising exceptions instead of issuing warnings, possibly introducing a "strict" mode to control the
    #  behavior
    try:
        description = get_census_version_description(census_version)  # raises
    except ValueError:
        raise ValueError(
            f'The "{census_version}" Census version is not valid. Use get_census_version_directory() to retrieve '
            f"available versions."
        ) from None

    if description.get("flags", {}).get("retracted", False):
        api_logger.warning(
            f"The \"{census_version}\" Census version has been retracted!\n{description['retraction']}."
            f'Use "stable" or "latest", or use get_census_version_directory() to retrieve valid versions.'
        )
    elif census_version == "stable":
        api_logger.info(
            f"The \"{census_version}\" release is currently {description['release_build']}. Specify "
            f"'census_version=\"{description['release_build']}\"' in future calls to open_soma() to ensure data "
            "consistency."
        )

    locator = _resolve_census_locator(description["soma"], selected_mirror)

    return _open_soma(locator, context)


def get_source_h5ad_uri(dataset_id: str, *, census_version: str = DEFAULT_CENSUS_VERSION) -> CensusLocator:
    """Open the named version of the census, and return the URI for the ``dataset_id``. This
    does not guarantee that the H5AD exists or is accessible to the user.

    Args:
        dataset_id:
            The ``dataset_id`` of interest.
        census_version:
            The census version. Defaults to `stable`.

    Returns:
        A :py:obj:`CensusLocator` object that contains the URI and optional S3 region for the source H5AD.

    Raises:
        KeyError: if either `dataset_id` or `census_version` do not exist.

    Lifecycle:
        maturing

    Examples:
        >>> cellxgene_census.get_source_h5ad_uri("cb5efdb0-f91c-4cbd-9ad4-9d4fa41c572d")
        {'uri': 's3://cellxgene-data-public/cell-census/2022-12-01/h5ads/cb5efdb0-f91c-4cbd-9ad4-9d4fa41c572d.h5ad',
        's3_region': 'us-west-2'}
    """
    description = get_census_version_description(census_version)  # raises

    # For h5ads, it makes sense to use the default mirror, since the artifacts themselves won't be mirrored
    mirrors = _get_census_mirrors()
    selected_mirror: CensusMirror = mirrors[mirrors["default"]]  # type: ignore
    census_locator = _resolve_census_locator(description["soma"], selected_mirror)

    census = _open_soma(census_locator)
    dataset = census["census_info"]["datasets"].read(value_filter=f"dataset_id == '{dataset_id}'").concat().to_pandas()
    if len(dataset) == 0:
        raise KeyError("Unknown dataset_id")

    locator = description["h5ads"].copy()
    h5ads_base_uri = locator["uri"]
    dataset_h5ad_path = dataset.dataset_h5ad_path.iloc[0]
    locator["uri"] = _uri_join(h5ads_base_uri, dataset_h5ad_path)
    return locator


def download_source_h5ad(dataset_id: str, to_path: str, *, census_version: str = DEFAULT_CENSUS_VERSION) -> None:
    """Download the source H5AD dataset, for the given `dataset_id`, to the user-specified
    file name.

    Args:
        dataset_id
            Fetch the source (original) H5AD associated with this `dataset_id`.
        to_path:
            The file name where the downloaded H5AD will be written.  Must not already exist.
        census_version:
            The census version name. Defaults to `stable`.

    Raises:
        ValueError: if the path already exists (i.e., will not overwrite
            an existing file), or is not a file.

    Lifecycle:
        maturing

    See Also:
        :func:`get_source_h5ad_uri`: Look up the location of the source H5AD.

    Examples:
        >>> download_source_h5ad("8e47ed12-c658-4252-b126-381df8d52a3d", to_path="/tmp/data.h5ad")
    """
    if os.path.exists(to_path):
        raise ValueError("Path exists - will not overwrite existing file.")
    if to_path.endswith("/"):
        raise ValueError("Specify to_path as a file name, not a directory name.")

    locator = get_source_h5ad_uri(dataset_id, census_version=census_version)
    protocol = urllib.parse.urlparse(locator["uri"]).scheme
    assert protocol == "s3"

    fs = s3fs.S3FileSystem(
        anon=True,
        cache_regions=True,
    )
    fs.get_file(locator["uri"], to_path)



import gc
import logging
import os
from contextlib import contextmanager
from datetime import timedelta
from math import ceil
from time import time
from typing import Any, Dict, Iterator, List, Optional, Sequence, Tuple

import numpy as np
import numpy.typing as npt
import pandas as pd
import psutil
import pyarrow as pa
import scipy
import tiledbsoma as soma
import torch
import torchdata.datapipes.iter as pipes
from attr import define
from numpy.random import Generator
from scipy import sparse
from sklearn.preprocessing import LabelEncoder
from torch import Tensor
from torch import distributed as dist
from torch.utils.data import DataLoader
from torch.utils.data.dataset import Dataset

from ... import get_default_soma_context
from ..util._eager_iter import _EagerIterator

pytorch_logger = logging.getLogger("cellxgene_census.experimental.pytorch")

# TODO: Rename to reflect the correct order of the Tensors within the tuple: (X, obs)
ObsAndXDatum = Tuple[Tensor, Tensor]
"""Return type of ``ExperimentDataPipe`` that pairs a Tensor of ``obs`` row(s) with a Tensor of ``X`` matrix row(s). 
The Tensors are rank 1 if ``batch_size`` is 1, otherwise the Tensors are rank 2."""


@define
class _SOMAChunk:
    """
    Return type of ``_ObsAndXSOMAIterator`` that pairs a chunk of ``obs`` rows with the respective rows from the ``X``
    matrix.

    Lifecycle:
        experimental
    """

    obs: pd.DataFrame
    X: scipy.sparse.spmatrix
    stats: "Stats"

    def __len__(self) -> int:
        return len(self.obs)


Encoders = Dict[str, LabelEncoder]
"""A dictionary of ``LabelEncoder``s keyed by the ``obs`` column name."""


@define
class Stats:
    """
    Statistics about the data retrieved by ``ExperimentDataPipe`` via SOMA API. This is useful for assessing the read
    throughput of SOMA data.

    Lifecycle:
        experimental
    """

    n_obs: int = 0
    """The total number of obs rows retrieved"""

    nnz: int = 0
    """The total number of values retrieved"""

    elapsed: int = 0
    """The total elapsed time in seconds for retrieving all batches"""

    n_soma_chunks: int = 0
    """The number of chunks retrieved"""

    def __str__(self) -> str:
        return f"{self.n_soma_chunks=}, {self.n_obs=}, {self.nnz=}, " f"elapsed={timedelta(seconds=self.elapsed)}"

    def __add__(self, other: "Stats") -> "Stats":
        self.n_obs += other.n_obs
        self.nnz += other.nnz
        self.elapsed += other.elapsed
        self.n_soma_chunks += other.n_soma_chunks
        return self


@contextmanager
def _open_experiment(
    uri: str,
    aws_region: Optional[str] = None,
) -> soma.Experiment:
    """Internal method for opening a SOMA ``Experiment`` as a context manager."""

    context = get_default_soma_context().replace(tiledb_config={"vfs.s3.region": aws_region} if aws_region else {})

    with soma.Experiment.open(uri, context=context) as exp:
        yield exp


class _ObsAndXSOMAIterator(Iterator[_SOMAChunk]):
    """Iterates the SOMA chunks of corresponding ``obs`` and ``X`` data. This is an internal class,
    not intended for public use."""

    X: soma.SparseNDArray
    """A handle to the full X data of the SOMA ``Experiment``"""

    obs_joinids_chunks_iter: Iterator[npt.NDArray[np.int64]]

    var_joinids: npt.NDArray[np.int64]
    """The ``var`` joinids to be retrieved from the SOMA ``Experiment``"""

    def __init__(
        self,
        obs: soma.DataFrame,
        X: soma.SparseNDArray,
        obs_column_names: Sequence[str],
        obs_joinids_chunked: List[npt.NDArray[np.int64]],
        var_joinids: npt.NDArray[np.int64],
        shuffle_rng: Optional[Generator] = None,
    ):
        self.obs = obs
        self.X = X
        self.obs_column_names = obs_column_names
        self.obs_joinids_chunks_iter = self._maybe_local_shuffle_obs_joinids(obs_joinids_chunked, shuffle_rng)
        self.var_joinids = var_joinids

    @staticmethod
    def _maybe_local_shuffle_obs_joinids(
        obs_joinids_chunked: List[npt.NDArray[np.int64]], shuffle_rng: Optional[Generator] = None
    ) -> Iterator[npt.NDArray[np.int64]]:
        return (
            shuffle_rng.permutation(obs_joinid_chunk) if shuffle_rng else obs_joinid_chunk
            for obs_joinid_chunk in obs_joinids_chunked
        )

    def __next__(self) -> _SOMAChunk:
        pytorch_logger.debug("Retrieving next SOMA chunk...")
        start_time = time()

        # If no more chunks to iterate through, raise StopIteration, as all iterators do when at end
        obs_joinids_chunk = next(self.obs_joinids_chunks_iter)

        obs_batch = (
            self.obs.read(
                coords=(obs_joinids_chunk,),
                column_names=self.obs_column_names,
            )
            .concat()
            .to_pandas()
            .set_index("soma_joinid")
        )
        assert obs_batch.shape[0] == obs_joinids_chunk.shape[0]

        # handle case of empty result (first batch has 0 rows)
        if len(obs_batch) == 0:
            raise StopIteration

        # reorder obs rows to match obs_joinids_chunk ordering, which may be shuffled
        obs_batch = obs_batch.reindex(obs_joinids_chunk, copy=False)

        # note: the `blockwise` call is employed for its ability to reindex the axes of the sparse matrix,
        # but the blockwise iteration feature is not used (block_size is set to retrieve the chunk as a single block)
        scipy_iter = (
            self.X.read(coords=(obs_joinids_chunk, self.var_joinids))
            .blockwise(axis=0, size=len(obs_joinids_chunk), eager=False)
            .scipy(compress=True)
        )
        X_batch, _ = next(scipy_iter)
        assert obs_batch.shape[0] == X_batch.shape[0]

        stats = Stats()
        stats.n_obs += X_batch.shape[0]
        stats.nnz += X_batch.nnz
        stats.elapsed += int(time() - start_time)
        stats.n_soma_chunks += 1

        pytorch_logger.debug(f"Retrieved SOMA chunk: {stats}")
        return _SOMAChunk(obs=obs_batch, X=X_batch, stats=stats)


def run_gc() -> Tuple[Tuple[Any, Any, Any], Tuple[Any, Any, Any]]:
    proc = psutil.Process(os.getpid())

    pre_gc = proc.memory_full_info(), psutil.virtual_memory(), psutil.swap_memory()
    gc.collect()
    post_gc = proc.memory_full_info(), psutil.virtual_memory(), psutil.swap_memory()

    pytorch_logger.debug(f"gc:  pre={pre_gc}")
    pytorch_logger.debug(f"gc: post={post_gc}")

    return pre_gc, post_gc


class _ObsAndXIterator(Iterator[ObsAndXDatum]):
    """
    Iterates through a set of ``obs`` and corresponding ``X`` rows, where the rows to be returned are specified by
    the ``obs_tables_iter`` argument. For the specified ``obs` rows, the corresponding ``X`` data is loaded and
    joined together. It is returned from this iterator as 2-tuples of ``X`` and obs Tensors.

    Internally manages the retrieval of data in SOMA-sized chunks, fetching the next chunk of SOMA data as needed.
    Supports fetching the data in an eager manner, where the next SOMA chunk is fetched while the current chunk is
    being read. This is an internal class, not intended for public use.
    """

    soma_chunk_iter: Iterator[_SOMAChunk]
    """The iterator for SOMA chunks of paired obs and X data"""

    soma_chunk: Optional[_SOMAChunk]
    """The current SOMA chunk of obs and X data"""

    i: int = -1
    """Index into current obs ``SOMA`` chunk"""

    def __init__(
        self,
        obs: soma.DataFrame,
        X: soma.SparseNDArray,
        obs_column_names: Sequence[str],
        obs_joinids_chunked: List[npt.NDArray[np.int64]],
        var_joinids: npt.NDArray[np.int64],
        batch_size: int,
        encoders: Dict[str, LabelEncoder],
        stats: Stats,
        return_sparse_X: bool,
        use_eager_fetch: bool,
        shuffle_rng: Optional[Generator] = None,
    ) -> None:
        self.soma_chunk_iter = _ObsAndXSOMAIterator(
            obs, X, obs_column_names, obs_joinids_chunked, var_joinids, shuffle_rng
        )
        if use_eager_fetch:
            self.soma_chunk_iter = _EagerIterator(self.soma_chunk_iter)
        self.soma_chunk = None
        self.var_joinids = var_joinids
        self.batch_size = batch_size
        self.return_sparse_X = return_sparse_X
        self.encoders = encoders
        self.stats = stats
        self.max_process_mem_usage_bytes = 0
        self.X_dtype = X.schema[2].type.to_pandas_dtype()

    def __next__(self) -> ObsAndXDatum:
        """Read the next torch batch, possibly across multiple soma chunks"""

        obs: pd.DataFrame = pd.DataFrame()
        X: sparse.csr_matrix = sparse.csr_matrix((0, len(self.var_joinids)), dtype=self.X_dtype)

        while len(obs) < self.batch_size:
            try:
                obs_partial, X_partial = self._read_partial_torch_batch(self.batch_size - len(obs))
                obs = pd.concat([obs, obs_partial], axis=0)
                X = sparse.vstack([X, X_partial])
            except StopIteration:
                break

        if len(obs) == 0:
            raise StopIteration

        obs_encoded = pd.DataFrame(
            data={"soma_joinid": obs.index}, columns=["soma_joinid"] + obs.columns.tolist(), dtype=np.int64
        )
        # TODO: Encode the entire SOMA chunk at once in _read_partial_torch_batch()
        for col, enc in self.encoders.items():
            obs_encoded[col] = enc.transform(obs[col])

        # `to_numpy()` avoids copying the numpy array data
        obs_tensor = torch.from_numpy(obs_encoded.to_numpy())

        if not self.return_sparse_X:
            X_tensor = torch.from_numpy(X.todense())
        else:
            coo = X.tocoo()

            X_tensor = torch.sparse_coo_tensor(
                # Note: The `np.array` seems unnecessary, but PyTorch warns bare array is "extremely slow"
                indices=torch.from_numpy(np.array([coo.row, coo.col])),
                values=coo.data,
                size=coo.shape,
            )

        if self.batch_size == 1:
            X_tensor = X_tensor[0]
            obs_tensor = obs_tensor[0]

        return X_tensor, obs_tensor

    def _read_partial_torch_batch(self, batch_size: int) -> ObsAndXDatum:
        """Reads a torch-size batch of data from the current SOMA chunk, returning a torch-size batch whose size may
        contain fewer rows than the requested ``batch_size``. This can happen when the remaining rows in the current
        SOMA chunk are fewer than the requested ``batch_size``."""

        if self.soma_chunk is None or not (0 <= self.i < len(self.soma_chunk)):
            # GC memory from previous soma_chunk
            self.soma_chunk = None
            mem_info = run_gc()
            self.max_process_mem_usage_bytes = max(self.max_process_mem_usage_bytes, mem_info[0][0].uss)

            self.soma_chunk: _SOMAChunk = next(self.soma_chunk_iter)
            self.stats += self.soma_chunk.stats
            self.i = 0

            pytorch_logger.debug(f"Retrieved SOMA chunk totals: {self.stats}")

        obs_batch = self.soma_chunk.obs
        X_batch = self.soma_chunk.X

        safe_batch_size = min(batch_size, len(obs_batch) - self.i)
        slice_ = slice(self.i, self.i + safe_batch_size)
        assert slice_.stop <= obs_batch.shape[0]

        obs_rows = obs_batch.iloc[slice_]
        assert obs_rows.index.is_unique
        assert safe_batch_size == obs_rows.shape[0]

        X_csr_scipy = X_batch[slice_]
        assert obs_rows.shape[0] == X_csr_scipy.shape[0]

        self.i += safe_batch_size

        return obs_rows, X_csr_scipy


class ExperimentDataPipe(pipes.IterDataPipe[Dataset[ObsAndXDatum]]):  # type: ignore
    """
    An iterable-style PyTorch ``DataPipe`` that reads ``obs`` and ``X`` data from a SOMA ``Experiment``, based upon the
    specified queries along the ``obs`` and ``var`` axes. Provides an iterator over these data when the object is
    passed to Python's built-in ``iter`` function:

    >>> for batch in iter(ExperimentDataPipe(...)):
            X_batch, y_batch = batch

    The ``batch_size`` parameter controls the number of rows of ``obs`` and ``X`` data that are returned in each
    iteration. If the ``batch_size`` is 1, then each Tensor will have rank 1:

    >>> (tensor([0., 0., 0., 0., 0., 1., 0., 0., 0.]),  # X data
         tensor([2415,    0,    0], dtype=torch.int64)) # obs data, encoded

    For larger ``batch_size`` values, the returned Tensors will have rank 2:

    >>> DataLoader(..., batch_size=3, ...):
        (tensor([[0., 0., 0., 0., 0., 1., 0., 0., 0.],     # X batch
                 [0., 0., 0., 0., 0., 0., 0., 0., 0.],
                 [0., 0., 0., 0., 0., 0., 0., 0., 0.]]),
         tensor([[2415,    0,    0],                       # obs batch
                 [2416,    0,    4],
                 [2417,    0,    3]], dtype=torch.int64))

    The ``return_sparse_X`` parameter controls whether the ``X`` data is returned as a dense or sparse Tensor.  If the
    model supports use of sparse Tensors, this will reduce memory usage.

    The ``obs_column_names`` parameter determines the data columns that are returned in the ``obs`` Tensor. The first
    element is always the ``soma_joinid`` of the ``obs`` DataFrame (or, equiavalently, the ``soma_dim_0`` of the ``X``
    matrix). The remaining elements are the ``obs`` columns specified by ``obs_column_names``, and string-typed
    columns are encoded as integer values. If needed, these values can be decoded by obtaining the encoder for a
    given ``obs`` column name and calling its ``inverse_transform`` method:

    >>> exp_data_pipe.obs_encoders["<obs_attr_name>"].inverse_transform(encoded_values)

    Lifecycle:
        experimental
    """

    _initialized: bool

    _obs_joinids: Optional[npt.NDArray[np.int64]]

    _var_joinids: Optional[npt.NDArray[np.int64]]

    _encoders: Optional[Encoders]

    _stats: Stats

    _shuffle_rng: Optional[Generator]

    # TODO: Consider adding another convenience method wrapper to construct this object whose signature is more closely
    #  aligned with get_anndata() params (i.e. "exploded" AxisQuery params).
    def __init__(
        self,
        experiment: soma.Experiment,
        measurement_name: str = "raw",
        X_name: str = "X",
        obs_query: Optional[soma.AxisQuery] = None,
        var_query: Optional[soma.AxisQuery] = None,
        obs_column_names: Sequence[str] = (),
        batch_size: int = 1,
        shuffle: bool = False,
        seed: Optional[int] = None,
        return_sparse_X: bool = False,
        soma_chunk_size: Optional[int] = None,
        use_eager_fetch: bool = True,
    ) -> None:
        """
        Construct a new ``ExperimentDataPipe``.

        Args:
            experiment:
                The SOMA ``Experiment`` from which to read data.
            measurement_name:
                The name of the SOMA ``Measurement`` to read. Defaults to "raw".
            X_name:
                The name of the X layer to read. Defaults to "X".
            obs_query:
                The query used to filter along the ``obs`` axis. If not specified, all ``obs`` and ``X`` data will
                be returned, which can be very large.
            var_query:
                The query used to filter along the ``var`` axis. If not specified, all ``var`` columns (genes/features)
                will be returned.
            obs_column_names:
                The names of the ``obs`` columns to return. The ``soma_joinid`` index "column" does not need to be
                specified and will always be returned. If not specified, only the ``soma_joinid`` will be returned.
            batch_size:
                The number of rows of ``obs`` and ``X`` data to return in each iteration. Defaults to 1. A value of 1
                will result in Tensors of rank 1 being returns (a single row); larger values will result in Tensors of
                rank 2 (multiple rows).
            shuffle:
                Whether to shuffle the ``obs`` and ``X`` data being returned. Defaults to ``False`` (no shuffling).
                For performance reasons, shuffling is performed in two steps: 1) a global shuffling, where contiguous
                rows are grouped into chunks and the order of the chunks is randomized, and then 2) a local
                shuffling, where the rows within each chunk are shuffled. Since ExperimentDataPipe must retrieve data
                in chunks (to keep memory requirements to a fixed size), global shuffling ensures that a given row in
                the shuffled result can originate from any position in the non-shuffled result ordering. If shuffling
                only occurred within each chunk (i.e. "local" shuffling), the first chunk's rows would always be
                returned first, the second chunk's rows would always be returned second, and so on. The chunk size is
                determined by the ``soma_chunk_size`` parameter. Note that rows within a chunk will maintain
                proximity, even after shuffling, so some experimentation may be required to ensure the shuffling is
                sufficient for the model training process. To this end, the ``soma_chunk_size`` can be treated as a
                hyperparameter that can be tuned.
            seed:
                The random seed used for shuffling. Defaults to ``None`` (no seed). This *must* be specified when using
                ``DistributedDataParallel`` to ensure data partitions are disjoint across worker processes.
            return_sparse_X:
                Controls whether the ``X`` data is returned as a dense or sparse Tensor. As ``X`` data is very sparse,
                setting this to ``True`` will reduce memory usage, if the model supports use of sparse Tensors. Defaults
                to ``False``, since sparse Tensors are still experimental in PyTorch.
            soma_chunk_size:
                The number of obs/X rows to retrieve when reading data from SOMA. This impacts two aspects of
                ``ExperimentDataPipe`` behavior: 1) The maximum memory utilization, with larger values providing
                better read performance, but also requiring more memory; 2) The granularity of the global shuffling
                step (see ``shuffle`` parameter for details). If not specified, the value is set to utilize ~1 GiB of
                RAM per SOMA chunk read, based upon the number of ``var`` columns (cells/features) being requested
                and assuming X data sparsity of 95%; the number of rows per chunk will depend on the number of
                ``var`` columns being read.
            use_eager_fetch:
                Fetch the next SOMA chunk of ``obs`` and ``X`` data immediately after a previously fetched SOMA chunk is made
                available for processing via the iterator. This allows network (or filesystem) requests to be made in
                parallel with client-side processing of the SOMA data, potentially improving overall performance at the
                cost of doubling memory utilization. Defaults to ``True``.
        Returns:
            The constructed ``ExperimentDataPipe``.

        Lifecycle:
            experimental
        """
        self.exp_uri = experiment.uri
        self.aws_region = experiment.context.tiledb_ctx.config().get("vfs.s3.region")
        self.measurement_name = measurement_name
        self.layer_name = X_name
        self.obs_query = obs_query
        self.var_query = var_query
        self.obs_column_names = obs_column_names
        self.batch_size = batch_size
        self.return_sparse_X = return_sparse_X
        self.soma_chunk_size = soma_chunk_size
        self.use_eager_fetch = use_eager_fetch
        self._stats = Stats()
        self._encoders = None
        self._obs_joinids = None
        self._var_joinids = None
        self._shuffle_rng = np.random.default_rng(seed) if shuffle else None
        self._initialized = False

        if "soma_joinid" not in self.obs_column_names:
            self.obs_column_names = ["soma_joinid", *self.obs_column_names]

    def _init(self) -> None:
        if self._initialized:
            return

        pytorch_logger.debug("Initializing ExperimentDataPipe")

        with _open_experiment(self.exp_uri, self.aws_region) as exp:
            query = exp.axis_query(
                measurement_name=self.measurement_name,
                obs_query=self.obs_query,
                var_query=self.var_query,
            )

            # The to_numpy() call is a workaround for a possible bug in TileDB-SOMA:
            # https://github.com/single-cell-data/TileDB-SOMA/issues/1456
            self._obs_joinids = query.obs_joinids().to_numpy()
            self._var_joinids = query.var_joinids().to_numpy()

            self._encoders = self._build_obs_encoders(query)

        self._initialized = True

    @staticmethod
    def _subset_ids_to_partition(
        ids_chunked: List[npt.NDArray[np.int64]], partition_index: int, num_partitions: int
    ) -> List[npt.NDArray[np.int64]]:
        """Returns a single partition of the obs_joinids_chunked (a 2D ndarray), based upon the current process's distributed rank and world
        size."""

        # subset to a single partition
        # typing does not reflect that is actually a List of 2D NDArrays
        partition_indices = np.array_split(range(len(ids_chunked)), num_partitions)
        partition = [ids_chunked[i] for i in partition_indices[partition_index]]

        if pytorch_logger.isEnabledFor(logging.DEBUG) and len(partition) > 0:
            pytorch_logger.debug(
                f"Process {os.getpid()} handling partition {partition_index + 1} of {num_partitions}, "
                f"partition_size={sum([len(chunk) for chunk in partition])}"
            )

        return partition

    @staticmethod
    def _compute_partitions(
        loader_partition: int, loader_partitions: int, dist_partition: int, num_dist_partitions: int
    ) -> Tuple[int, int]:
        # NOTE: Can alternately use a `worker_init_fn` to split among workers split workload
        total_partitions = num_dist_partitions * loader_partitions
        partition = dist_partition * loader_partitions + loader_partition
        return partition, total_partitions

    def __iter__(self) -> Iterator[ObsAndXDatum]:
        self._init()
        assert self._obs_joinids is not None
        assert self._var_joinids is not None

        if self.soma_chunk_size is None:
            # set soma_chunk_size to utilize ~1 GiB of RAM per SOMA chunk; assumes 95% X data sparsity, 8 bytes for the
            # X value and 8 bytes for the sparse matrix indices, and a 100% working memory overhead (2x).
            X_row_memory_size = 0.05 * len(self._var_joinids) * 8 * 3 * 2
            self.soma_chunk_size = int((1 * 1024**3) / X_row_memory_size)
        pytorch_logger.debug(f"Using {self.soma_chunk_size=}")

        if (
            self.return_sparse_X
            and torch.utils.data.get_worker_info()
            and torch.utils.data.get_worker_info().num_workers > 0
        ):
            raise NotImplementedError(
                "torch does not work with sparse tensors in multi-processing mode "
                "(see https://github.com/pytorch/pytorch/issues/20248)"
            )

        # chunk the obs joinids into batches of size soma_chunk_size
        obs_joinids_chunked = self._chunk_ids(self._obs_joinids, self.soma_chunk_size)

        # globally shuffle the chunks, if requested
        if self._shuffle_rng:
            self._shuffle_rng.shuffle(obs_joinids_chunked)

        # subset to a single partition, as needed for distributed training and multi-processing datat loading
        worker_info = torch.utils.data.get_worker_info()
        partition, partitions = self._compute_partitions(
            loader_partition=worker_info.id if worker_info else 0,
            loader_partitions=worker_info.num_workers if worker_info else 1,
            dist_partition=dist.get_rank() if dist.is_initialized() else 0,
            num_dist_partitions=dist.get_world_size() if dist.is_initialized() else 1,
        )
        obs_joinids_chunked_partition: List[npt.NDArray[np.int64]] = self._subset_ids_to_partition(
            obs_joinids_chunked, partition, partitions
        )

        with _open_experiment(self.exp_uri, self.aws_region) as exp:
            obs_and_x_iter = _ObsAndXIterator(
                obs=exp.obs,
                X=exp.ms[self.measurement_name].X[self.layer_name],
                obs_column_names=self.obs_column_names,
                obs_joinids_chunked=obs_joinids_chunked_partition,
                var_joinids=self._var_joinids,
                batch_size=self.batch_size,
                encoders=self.obs_encoders,
                stats=self._stats,
                return_sparse_X=self.return_sparse_X,
                use_eager_fetch=self.use_eager_fetch,
                shuffle_rng=self._shuffle_rng,
            )

            for datum_ in obs_and_x_iter:
                yield datum_

            pytorch_logger.debug(
                "max process memory usage=" f"{obs_and_x_iter.max_process_mem_usage_bytes / (1024 ** 3):.3f} GiB"
            )

    @staticmethod
    def _chunk_ids(ids: npt.NDArray[np.int64], chunk_size: int) -> List[npt.NDArray[np.int64]]:
        num_chunks = max(1, ceil(len(ids) / chunk_size))
        pytorch_logger.debug(f"Shuffling {len(ids)} obs joinids into {num_chunks} chunks of {chunk_size}")
        return np.array_split(ids, num_chunks)

    def __len__(self) -> int:
        self._init()
        assert self._obs_joinids is not None

        return len(self._obs_joinids)

    def __getitem__(self, index: int) -> ObsAndXDatum:
        raise NotImplementedError("IterDataPipe can only be iterated")

    def _build_obs_encoders(self, query: soma.ExperimentAxisQuery) -> Encoders:
        pytorch_logger.debug("Initializing encoders")

        obs = query.obs(column_names=self.obs_column_names).concat()
        encoders = {}
        for col in self.obs_column_names:
            if obs[col].type in (pa.string(), pa.large_string()):
                enc = LabelEncoder()
                enc.fit(obs[col].combine_chunks().unique())
                encoders[col] = enc

        return encoders

    # TODO: This does not work in multiprocessing mode, as child process's stats are not collected
    def stats(self) -> Stats:
        """
        Get data loading stats for this ``ExperimentDataPipe``.

        Returns:
            The ``Stats`` object for this ``ExperimentDataPipe``.

        Lifecycle:
            experimental
        """
        return self._stats

    @property
    def shape(self) -> Tuple[int, int]:
        """
        Get the shape of the data that will be returned by this ``ExperimentDataPipe``. This is the number of
        obs (cell) and var (feature) counts in the returned data. If used in multiprocessing mode
        (i.e. DataLoader instantiated with num_workers > 0), the obs (cell) count will reflect the size of the
        partition of the data assigned to the active process.

        Returns:
            A 2-tuple of ``int``s, for obs and var counts, respectively.

        Lifecycle:
            experimental
        """
        self._init()
        assert self._obs_joinids is not None
        assert self._var_joinids is not None

        return len(self._obs_joinids), len(self._var_joinids)

    @property
    def obs_encoders(self) -> Encoders:
        """
        Returns a dictionary of ``sklearn.preprocessing.LabelEncoder`` objects, keyed on ``obs`` column names,
        which were used to encode the ``obs`` column values. These encoders can be used to decode the encoded values as
        follows:

        >>> exp_data_pipe.obs_encoders["<obs_attr_name>"].inverse_transform(encoded_values)

        See https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.LabelEncoder.html.

        Returns:
            A ``Dict[str, LabelEncoder]``, mapping column names to ``LabelEncoder``s.
        """
        self._init()
        assert self._encoders is not None

        return self._encoders


# Note: must be a top-level function (and not a lambda), to play nice with multiprocessing pickling
def _collate_noop(x: Any) -> Any:
    return x


# TODO: Move into somacore.ExperimentAxisQuery
def experiment_dataloader(
    datapipe: pipes.IterDataPipe,
    num_workers: int = 0,
    **dataloader_kwargs: Any,
) -> DataLoader:
    """
    Factory method for PyTorch ``DataLoader``. This method can be used to safely instantiate a
    ``DataLoader`` that works with the ``ExperimentDataPipe``, since some of the ``DataLoader`` constructor params
    are not applicable when using a ``IterDataPipe`` (``shuffle``, ``batch_size``, ``sampler``, ``batch_sampler``,
    ``collate_fn``).

    Args:
        datapipe:
            A PyTorch ``IterDataPipe``, which can be an ``ExperimentDataPipe`` or any other ``IterDataPipe`` that has
            been chained to the ``ExperimentDataPipe``.
        num_workers:
            Number of worker processes to use for data loading. If 0, data will be loaded in the main process.
        **dataloader_kwargs:
            Additional keyword arguments to pass to the ``torch.utils.data.DataLoader`` constructor,
            except for ``shuffle``, ``batch_size``, ``sampler``, ``batch_sampler``, and ``collate_fn``, which are not
            supported when using ``ExperimentDataPipe``.
            See https://pytorch.org/docs/stable/data.html#torch.utils.data.DataLoader.

    Returns:
        A ``torch.utils.data.DataLoader``.

    Raises:
        ValueError: if any of the ``shuffle``, ``batch_size``, ``sampler``, ``batch_sampler``, or ``collate_fn`` params
        are passed as keyword arguments.


    Lifecycle:
        experimental
    """

    unsupported_dataloader_args = ["shuffle", "batch_size", "sampler", "batch_sampler", "collate_fn"]
    if set(unsupported_dataloader_args).intersection(dataloader_kwargs.keys()):
        raise ValueError(f"The {','.join(unsupported_dataloader_args)} DataLoader params are not supported")

    if num_workers > 0:
        _init_multiprocessing()

    return DataLoader(
        datapipe,
        batch_size=None,  # batching is handled by our ExperimentDataPipe
        num_workers=num_workers,
        # avoid use of default collator, which adds an extra (3rd) dimension to the tensor batches
        collate_fn=_collate_noop,
        # shuffling is handled by our ExperimentDataPipe
        shuffle=False,
        **dataloader_kwargs,
    )


def _init_multiprocessing() -> None:
    """Ensures use of "spawn" for starting child processes with multiprocessing.
    Forked processes are known to be problematic:
      https://pytorch.org/docs/stable/notes/multiprocessing.html#avoiding-and-fighting-deadlocks
    Also, CUDA does not support forked child processes:
      https://pytorch.org/docs/stable/notes/multiprocessing.html#cuda-in-multiprocessing

    """
    torch.multiprocessing.set_start_method("fork", force=True)
    orig_start_method = torch.multiprocessing.get_start_method()
    if orig_start_method != "spawn":
        if orig_start_method:
            pytorch_logger.warning(
                "switching torch multiprocessing start method from "
                f'"{torch.multiprocessing.get_start_method()}" to "spawn"'
            )
        torch.multiprocessing.set_start_method("spawn", force=True)



# Copyright (c) 2022-2023 Chan Zuckerberg Initiative Foundation
#
# Licensed under the MIT License.

"""
Methods to support simplifed access to community contributed embeddings.
"""
from __future__ import annotations

import json
import warnings
from typing import Any, Dict, Optional, Union, cast

import numpy as np
import numpy.typing as npt
import pandas as pd
import pyarrow as pa
import tiledbsoma as soma

from .._open import get_default_soma_context, open_soma
from .._release_directory import get_census_version_directory


def get_embedding_metadata(
    embedding_uri: str, context: Optional[soma.options.SOMATileDBContext] = None
) -> Dict[str, Any]:
    """
    Read embedding metadata and return as a Python dict.

    Args:
        embedding_uri:
            The embedding URI
        context:
            A custom :class:`SOMATileDBContext` which will be used to open the SOMA object.
            Optional, defaults to None.

    Returns:
        A Python dictionary containing metadata describing the embedding.

    Examples:
        >>> get_experiment_metadata(uri)

    """

    # Allow the user to override context for exceptional cases (e.g. the aws region)
    context = context or get_default_soma_context()

    with soma.open(embedding_uri, context=context) as E:
        # read embedding metadata and decode the JSON-encoded string
        embedding_metadata = json.loads(E.metadata["CxG_embedding_info"])
        assert isinstance(embedding_metadata, dict)

    return cast(Dict[str, Any], embedding_metadata)


def get_embedding(
    census_version: str,
    embedding_uri: str,
    obs_soma_joinids: Union[npt.NDArray[np.int64], pa.Array],
    context: Optional[soma.options.SOMATileDBContext] = None,
) -> npt.NDArray[np.float32]:
    """
    Read cell (obs) embeddings and return as a dense NumPy ndarray. Any cells without
    an embedding will return NaN values.

    Args:
        census_version:
            The Census version tag, e.g., "2023-12-15". Used to verify that the contents of
            the embedding contain embedded cells from the same Census version.
        embedding_uri:
            The URI containing the embedding data.
        obs_soma_joinids:
            The slice of the embedding to fetch and return.
        context:
            A custom :class:`SOMATileDBContext` which will be used to open the SOMA object.
            Optional, defaults to None.

    Returns:
        A :class:`numpy.ndarray` containing the embeddings. Embeddings are positionally
        indexed by the obs_soma_joinids. In other words, the cell identified by
        `obs_soma_joinids[i]` corresponds to the `ith` position in the returned ndarray.

    Raises:
        ValueError: if the Census and embedding are mismatched.

    Lifecycle:
        experimental

    Examples:
        >>> obs_somaids_to_fetch = np.array([10,11], dtype=np.int64)
        >>> emb = cellxgene_census.experimental.get_embedding('2023-10-23', embedding_uri, obs_somaids_to_fetch)
        >>> emb.shape
        (2, 200)
        >>> emb[:, 0:4]
        array([[ 0.02954102,  1.0390625 , -0.14550781, -0.40820312],
            [-0.00224304,  1.265625  ,  0.05883789, -0.7890625 ]],
            dtype=float32)

    """

    if isinstance(obs_soma_joinids, (pa.Array, pa.ChunkedArray)):
        obs_soma_joinids = obs_soma_joinids.to_numpy()
    if obs_soma_joinids.dtype != np.int64:
        raise TypeError("obs_soma_joinids must be array of int64")

    # Allow the user to override context for exceptional cases (e.g. the aws region)
    context = context or get_default_soma_context()

    # Attempt to resolve census version aliases
    census_directory = get_census_version_directory()
    resolved_census_version = census_directory.get(census_version, None)

    with soma.open(embedding_uri, context=context) as E:
        embedding_metadata = json.loads(E.metadata["CxG_embedding_info"])

        if resolved_census_version is None:
            warnings.warn(
                "Unable to determine Census version - skipping validation of Census and embedding version.",
                stacklevel=1,
            )
        elif resolved_census_version != census_directory.get(embedding_metadata["census_version"], None):
            raise ValueError("Census and embedding mismatch - census_version not equal")

        with open_soma(census_version=census_version, context=context) as census:
            experiment_name = embedding_metadata["experiment_name"]
            if experiment_name not in census["census_data"]:
                raise ValueError("Census and embedding mismatch - experiment_name does not exist")
            measurement_name = embedding_metadata["measurement_name"]
            if measurement_name not in census["census_data"][experiment_name].ms:
                raise ValueError("Census and embedding mismatch - measurement_name does not exist")

        embedding_shape = (len(obs_soma_joinids), E.shape[1])
        embedding = np.full(embedding_shape, np.NaN, dtype=np.float32, order="C")

        obs_indexer = pd.Index(obs_soma_joinids)
        for tbl in E.read(coords=(obs_soma_joinids,)).tables():
            obs_idx = obs_indexer.get_indexer(tbl.column("soma_dim_0").to_numpy())  # type: ignore[no-untyped-call]
            feat_idx = tbl.column("soma_dim_1").to_numpy()
            emb = tbl.column("soma_data")

            indices = obs_idx * E.shape[1] + feat_idx
            np.put(embedding.reshape(-1), indices, emb)

    return embedding



"""
Compatibility tests between the installed verison of cellxgene-census and
a named LTS release. Primarilly intended to be driven by a periodic GHA.

Where there are known and accepted incompatibilities, use `pytest.skip`
to codify them.
"""

from __future__ import annotations

from collections import deque
from typing import Iterator, Literal, Optional, Sequence, Union, get_args

import pyarrow as pa
import pytest
import tiledbsoma as soma

import cellxgene_census

SOMATypeNames = Literal[
    "SOMACollection", "SOMAExperiment", "SOMAMeasurement", "SOMADataFrame", "SOMASparseNDArray", "SOMADenseNDArray"
]
CollectionTypeNames = ["SOMACollection", "SOMAExperiment", "SOMAMeasurement"]

SOMATypes = Union[
    soma.Collection, soma.DataFrame, soma.SparseNDArray, soma.DenseNDArray, soma.Experiment, soma.Measurement
]


def walk_census(
    census: soma.Collection, filter_types: Optional[Sequence[SOMATypeNames]] = None
) -> Iterator[tuple[str, SOMATypes]]:
    assert census.soma_type == "SOMACollection"
    filter_types = filter_types or get_args(SOMATypeNames)
    items_to_check = deque([("census", census)])
    while items_to_check:
        key, val = items_to_check.popleft()
        if val.soma_type in CollectionTypeNames:
            items_to_check.extend(val.items())

        if val.soma_type not in filter_types:
            continue

        yield key, val


@pytest.mark.lts_compat_check
def test_open(census_version: str) -> None:
    """
    Verify we can open and walk the collections, get metadata and read schema on non-collections
    """

    with cellxgene_census.open_soma(census_version=census_version) as census:
        for name, item in walk_census(census):
            assert name
            assert list(item.metadata)
            if item.soma_type not in CollectionTypeNames:
                assert isinstance(item.schema, pa.Schema)


@pytest.mark.lts_compat_check
def test_read_dataframe(census_version: str) -> None:
    """
    Verify we can read at least one row of dataframes
    """
    with cellxgene_census.open_soma(census_version=census_version) as census:
        for name, sdf in walk_census(census, filter_types=["SOMADataFrame"]):
            assert name
            # the Census should have no zero-length DataFrames
            assert len(sdf)
            df = sdf.read(coords=([0],)).concat().to_pandas()
            assert len(df) == 1
            assert df.shape == (1, len(sdf.keys()))


@pytest.mark.lts_compat_check
def test_read_arrays(census_version: str) -> None:
    """
    Verify we can read from NDArray
    """
    with cellxgene_census.open_soma(census_version=census_version) as census:
        for name, sarr in walk_census(census, filter_types=["SOMASparseNDArray", "SOMADenseNDArray"]):
            assert name
            assert isinstance(sarr.shape, tuple)

            # There are currently no Census schema versions using DenseNDArray

            assert sarr.soma_type == "SOMASparseNDArray"
            assert sarr.nnz
            tbl = sarr.read(coords=(slice(100),)).tables().concat()
            assert len(tbl) > 0



import uuid
from abc import ABC, abstractmethod
from typing import Any, Dict, Generator, Optional

import scipy.sparse
from datasets import Dataset
from tiledbsoma import Experiment, ExperimentAxisQuery


class CellDatasetBuilder(ExperimentAxisQuery[Experiment], ABC):  # type: ignore
    """
    Abstract base class for methods to process CELLxGENE Census ExperimentAxisQuery
    results into a Hugging Face Dataset in which each item represents one cell.
    Subclasses implement the `cell_item()` method to process each row of an X layer
    into a Dataset item, and may also override `__init__()` and context `__enter__()`
    to perform any necessary preprocessing.

    The base class inherits ExperimentAxisQuery, so typical usage would be:

    ```
    import cellxgene_census
    import tiledbsoma
    from cellxgene_census.experimental.ml import GeneformerTokenizer

    with cellxgene_census.open_soma() as census:
        with SubclassOfCellDatasetBuilder(
            census["census_data"]["homo_sapiens"],
            obs_query=tilebsoma.AxisQuery(...),  # define some subset of Census cells
            ... # other ExperimentAxisQuery parameters e.g. var_query
        ) as builder:
            dataset = builder.build()
    ```
    """

    def __init__(
        self,
        experiment: Experiment,
        measurement_name: str = "RNA",
        layer_name: str = "raw",
        *,
        block_size: Optional[int] = None,
        **kwargs: Any,
    ):
        """
        Initialize the CellDatasetBuilder to process the results of a Census
        ExperimentAxisQuery.

        - `experiment`: Census Experiment to be queried.
        - `measurement_name`: Measurement in the experiment, default "RNA".
        - `layer_name`: Name of the X layer to process, default "raw".
        - `block_size`: Number of cells to process in-memory at once. If unspecified,
           `tiledbsoma.SparseNDArrayRead.blockwise()` will select a default.
        - `kwargs`: passed through to `ExperimentAxisQuery()`, especially `obs_query`
           and `var_query`.
        """
        super().__init__(experiment, measurement_name, **kwargs)
        self.layer_name = layer_name
        self.block_size = block_size

    def build(self, from_generator_kwargs: Optional[Dict[str, Any]] = None) -> Dataset:
        """
        Build the dataset from query results

        - `from_generator_kwargs`: kwargs passed through to `Dataset.from_generator()`
        """

        def gen() -> Generator[Dict[str, Any], None, None]:
            for Xblock, (block_cell_joinids, _) in (
                self.X(self.layer_name).blockwise(axis=0, reindex_disable_on_axis=[1], size=self.block_size).scipy()
            ):
                assert isinstance(Xblock, scipy.sparse.csr_matrix)
                for i, cell_joinid in enumerate(block_cell_joinids):
                    yield self.cell_item(cell_joinid, Xblock.getrow(i))

        return Dataset.from_generator(_DatasetGeneratorPickleHack(gen), **(from_generator_kwargs or {}))

    @abstractmethod
    def cell_item(self, cell_joinid: int, Xrow: scipy.sparse.csr_matrix) -> Dict[str, Any]:
        """
        Abstract method to process the X row for one cell into a Dataset item.

        - `cell_joinid`: The cell `soma_joinid`.
        - `Xrow`: The `X` row for this cell. This csr_matrix has a single row 0, equal
          to the `cell_joinid` row of the full `X` layer matrix.
        """
        ...


class _DatasetGeneratorPickleHack:
    """
    SEE: https://github.com/huggingface/datasets/issues/6194
    """

    def __init__(self, generator: Any, generator_id: Optional[str] = None) -> None:
        self.generator = generator
        self.generator_id = generator_id if generator_id is not None else str(uuid.uuid4())

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self.generator(*args, **kwargs)

    def __reduce__(self) -> Any:
        return (_DatasetGeneratorPickleHack_raise, (self.generator_id,))


def _DatasetGeneratorPickleHack_raise(*args: Any, **kwargs: Any) -> None:
    raise AssertionError("cannot actually unpickle _DatasetGeneratorPickleHack!")



import datasets
import pytest
import tiledbsoma
from py.path import local as Path
from scipy.stats import spearmanr

import cellxgene_census

try:
    from geneformer import TranscriptomeTokenizer

    from cellxgene_census.experimental.ml.huggingface import GeneformerTokenizer
except ImportError:
    # this should only occur when not running `experimental`-marked tests
    pass


CENSUS_VERSION_FOR_GENEFORMER_TESTS = "2023-12-15"


@pytest.mark.experimental
@pytest.mark.live_corpus
def test_GeneformerTokenizer_correctness(tmpdir: Path) -> None:
    """
    Test that GeneformerTokenizer produces the same token sequences as the original
    geneformer.TranscriptomeTokenizer (modulo a small tolerance on Spearman rank correlation)
    """
    # causes deterministic selection of roughly 1,000 cells:
    MODULUS = 32768
    # minimum Spearman rank correlation to consider token sequences effectively identical; this
    # allows for rare, slight differences in token sequences possibly arising from unstable sorting
    # and/or minor numerical precision differences in lowly-expressed genes.
    RHO_THRESHOLD = 0.99
    # notwithstanding RHO_THRESHOLD, we'll check that almost all token sequences are -exactly-
    # identical.
    EXACT_THRESHOLD = 0.98

    with cellxgene_census.open_soma(census_version=CENSUS_VERSION_FOR_GENEFORMER_TESTS) as census:
        human = census["census_data"]["homo_sapiens"]
        # read obs dataframe to get soma_joinids of all primary cells
        obs_df = (
            human.obs.read(column_names=["soma_joinid"], value_filter="is_primary_data == True").concat().to_pandas()
        )
        # select those with soma_joinid == 0 (mod MODULUS)
        cell_ids = [it for it in obs_df["soma_joinid"].tolist() if it % MODULUS == 0]

        # run our GeneformerTokenizer on them
        with GeneformerTokenizer(
            human,
            obs_query=tiledbsoma.AxisQuery(coords=(cell_ids,)),
        ) as tokenizer:
            test_tokens = [it["input_ids"] for it in tokenizer.build()]

        # write h5ad for use with geneformer.TranscriptomeTokenizer
        ad = cellxgene_census.get_anndata(
            census,
            "homo_sapiens",
            X_name="raw",
            obs_coords=cell_ids,
            column_names=tiledbsoma.AxisColumnNames(var=["feature_id"]),
        )
        ad.var.rename(columns={"feature_id": "ensembl_id"}, inplace=True)
        ad.obs["n_counts"] = ad.X.sum(axis=1)
        h5ad_dir = tmpdir.join("h5ad")
        h5ad_dir.mkdir()
        ad.write_h5ad(h5ad_dir.join("tokenizeme.h5ad"))
        # run geneformer.TranscriptomeTokenizer to get "true" tokenizations
        # see: https://huggingface.co/ctheodoris/Geneformer/blob/main/geneformer/tokenizer.py
        TranscriptomeTokenizer({}).tokenize_data(h5ad_dir, tmpdir, "tk", file_format="h5ad")
        true_tokens = [it["input_ids"] for it in datasets.load_from_disk(tmpdir.join("tk.dataset"))]

        # check GeneformerTokenizer sequences against geneformer.TranscriptomeTokenizer's
        assert len(test_tokens) == len(cell_ids)
        assert len(true_tokens) == len(cell_ids)
        identical = 0
        for i, cell_id in enumerate(cell_ids):
            assert len(test_tokens[i]) == len(true_tokens[i])
            rho, _ = spearmanr(test_tokens[i], true_tokens[i])
            if rho < RHO_THRESHOLD:
                # token sequences are too dissimilar; assert exact identity so that pytest -vv will
                # show the complete diff:
                assert (
                    test_tokens[i] == true_tokens[i]
                ), f"Discrepant token sequences for cell soma_joinid={cell_id}; Spearman rho={rho}"
            elif test_tokens[i] == true_tokens[i]:
                identical += 1
        assert identical / len(cell_ids) >= EXACT_THRESHOLD


@pytest.mark.experimental
@pytest.mark.live_corpus
def test_GeneformerTokenizer_docstring_example() -> None:
    with cellxgene_census.open_soma(census_version=CENSUS_VERSION_FOR_GENEFORMER_TESTS) as census:
        with GeneformerTokenizer(
            census["census_data"]["homo_sapiens"],
            # set obs_query to define some subset of Census cells:
            obs_query=tiledbsoma.AxisQuery(value_filter="is_primary_data == True and tissue_general == 'tongue'"),
            obs_attributes=(
                "soma_joinid",
                "cell_type_ontology_term_id",
            ),
        ) as tokenizer:
            dataset = tokenizer.build()
            assert len(dataset) == 15020
            assert sum(it.length for it in dataset.to_pandas().itertuples()) == 27798388



from __future__ import annotations

from typing import Any, Dict, Optional, Union

import numpy as np
import pandas as pd
import pytest
import scanpy as sc
import tiledbsoma as soma

import cellxgene_census
from cellxgene_census.experimental.pp import (
    get_highly_variable_genes,
    highly_variable_genes,
)


@pytest.mark.experimental
@pytest.mark.live_corpus
@pytest.mark.parametrize(
    "experiment_name,obs_value_filter",
    [
        (
            "mus_musculus",
            'is_primary_data == True and tissue_general == "liver"',
        ),
        pytest.param(
            "mus_musculus",
            'is_primary_data == True and tissue_general == "skin of body"',
            marks=pytest.mark.expensive,
        ),
        pytest.param(
            "mus_musculus",
            'is_primary_data == True and tissue_general in ["heart", "lung"]',
            marks=pytest.mark.expensive,
        ),
        pytest.param(
            "mus_musculus",
            'is_primary_data == True and assay == "Smart-seq2"',
            marks=pytest.mark.expensive,
        ),
    ],
)
@pytest.mark.parametrize("n_top_genes", (50, 500))
@pytest.mark.parametrize(
    "batch_key",
    (
        None,
        "dataset_id",
        ["suspension_type", "assay_ontology_term_id"],
        pytest.param(
            ("suspension_type", "assay_ontology_term_id"),
            marks=pytest.mark.expensive,
        ),
    ),
)
@pytest.mark.parametrize(
    "span",
    (
        pytest.param(None, marks=pytest.mark.expensive),
        0.5,
    ),
)
@pytest.mark.parametrize(
    "version",
    (
        "latest",
        pytest.param("stable", marks=pytest.mark.expensive),
    ),
)
def test_hvg_vs_scanpy(
    n_top_genes: int,
    obs_value_filter: str,
    version: str,
    experiment_name: str,
    batch_key: Optional[Union[str, tuple[str], list[str]]],
    span: float,
    small_mem_context: soma.SOMATileDBContext,
) -> None:
    """Compare results with ScanPy on a couple of simple tests."""

    kwargs: Dict[str, Any] = dict(
        n_top_genes=n_top_genes,
        batch_key=batch_key,
        flavor="seurat_v3",
    )
    if span is not None:
        kwargs["span"] = span

    with cellxgene_census.open_soma(census_version=version, context=small_mem_context) as census:
        # Get the highly variable genes
        with census["census_data"][experiment_name].axis_query(
            measurement_name="RNA",
            obs_query=soma.AxisQuery(value_filter=obs_value_filter),
        ) as query:
            hvg = highly_variable_genes(query, **kwargs)
            adata = query.to_anndata(X_name="raw")

    if isinstance(batch_key, list) or isinstance(batch_key, tuple):
        # ScanPy only accepts a single column for a batch key, so create it and use it
        assert "the_batch_key" not in adata.obs.columns
        adata.obs["the_batch_key"] = (
            adata.obs[list(batch_key)].astype(str)[batch_key[0]].str.cat(adata.obs[list(batch_key[1:])])
        )
        kwargs["batch_key"] = "the_batch_key"

    try:
        scanpy_hvg = sc.pp.highly_variable_genes(adata, inplace=False, **kwargs)
    except (ZeroDivisionError, ValueError):
        # There are test cases where ScanPy will fail, rendering this "compare vs scanpy"
        # test moot. The known cases involve overly partitioned data that results in batches
        # with a very small number of samples (which manifest as a divide by zero error).
        # In these known cases, go ahead and perform the HVG (above), but skip the compare
        # assertions below.
        pytest.skip("ScanPy generated an error, likely due to batches with 1 sample")

    scanpy_hvg.index.name = "soma_joinid"
    scanpy_hvg.index = scanpy_hvg.index.astype(int)
    assert len(scanpy_hvg) == len(hvg)
    assert all(scanpy_hvg.keys() == hvg.keys())

    assert (hvg.index == scanpy_hvg.index).all()
    assert np.allclose(
        hvg.means.to_numpy(),
        scanpy_hvg.means.to_numpy(),
        atol=1e-5,
        rtol=1e-2,
        equal_nan=True,
    )
    assert np.allclose(
        hvg.variances.to_numpy(),
        scanpy_hvg.variances.to_numpy(),
        atol=1e-5,
        rtol=1e-2,
        equal_nan=True,
    )
    assert np.allclose(
        hvg.variances_norm.to_numpy(),
        scanpy_hvg.variances_norm.to_numpy(),
        atol=1e-5,
        rtol=1e-2,
        equal_nan=True,
    )

    # Online calculation of normalized variance will differ slightly from ScanPy's calculation,
    # so look for rank of HVGs to be close, but not identical.  Don't worry about the non-HVGs
    # (which will differ more as you get to the long tail).  This test just looks for the average
    # rank distance to be small.
    assert (
        (scanpy_hvg[scanpy_hvg.highly_variable].highly_variable_rank - hvg[hvg.highly_variable].highly_variable_rank)
        .abs()
        .sum()
        / n_top_genes
    ) < 0.05

    # Ranking will also have some noise, so check that ranking is close in the highly variable subset
    scanpy_rank = scanpy_hvg.highly_variable_rank.copy()
    hvg_rank = hvg.highly_variable_rank.copy()
    hvg_rank[pd.isna(hvg_rank)] = n_top_genes
    scanpy_rank[pd.isna(scanpy_rank)] = n_top_genes
    rank_diff = (hvg_rank - scanpy_rank)[hvg.highly_variable]
    # +/- 5 in ranking, choosen arbitrarily
    assert rank_diff.min() >= -5 and rank_diff.max() <= 5

    if "highly_variable_nbatches" in scanpy_hvg.keys() or "highly_variable_nbatches" in hvg.keys():
        # Also subject to noise, so look for "close" match
        nbatches_diff = hvg.highly_variable_nbatches - scanpy_hvg.highly_variable_nbatches
        assert nbatches_diff.min() >= -2 and nbatches_diff.max() <= 2

    assert (hvg.highly_variable == scanpy_hvg.highly_variable).all()


@pytest.mark.experimental
@pytest.mark.live_corpus
@pytest.mark.parametrize(
    "experiment_name,organism,obs_value_filter,batch_key,obs_coords",
    [
        (
            "mus_musculus",
            "Mus musculus",
            'tissue_general == "liver" and is_primary_data == True',
            None,
            None,
        ),
        (
            "mus_musculus",
            "Mus musculus",
            'is_primary_data == True and tissue_general == "heart"',
            "dataset_id",
            None,
        ),
        (
            "mus_musculus",
            "Mus musculus",
            'is_primary_data == True and tissue_general == "heart"',
            ["suspension_type", "assay_ontology_term_id"],
            None,
        ),
        pytest.param(
            "mus_musculus",
            "Mus musculus",
            "is_primary_data == True",
            "dataset_id",
            slice(750_000, 1_000_000),
            marks=pytest.mark.expensive,
        ),
        pytest.param(
            "homo_sapiens",
            "Homo sapiens",
            "is_primary_data == True",
            "dataset_id",
            slice(1_000_000, 4_000_000),
            marks=pytest.mark.expensive,
        ),
    ],
)
def test_get_highly_variable_genes(
    organism: str,
    experiment_name: str,
    obs_value_filter: str,
    batch_key: str,
    small_mem_context: soma.SOMATileDBContext,
    obs_coords: Optional[slice],
) -> None:
    with cellxgene_census.open_soma(census_version="stable", context=small_mem_context) as census:
        hvg = get_highly_variable_genes(
            census,
            organism=organism,
            obs_value_filter=obs_value_filter,
            n_top_genes=1000,
            batch_key=batch_key,
            obs_coords=obs_coords,
        )
        n_vars = census["census_data"][experiment_name].ms["RNA"].var.count

    assert isinstance(hvg, pd.DataFrame)
    assert len(hvg) == n_vars
    assert len(hvg[hvg.highly_variable]) == 1000


@pytest.mark.experimental
def test_hvg_error_cases(small_mem_context: soma.SOMATileDBContext) -> None:
    with cellxgene_census.open_soma(census_version="stable", context=small_mem_context) as census:
        with census["census_data"]["mus_musculus"].axis_query(measurement_name="RNA") as query:
            # Only flavor="seurat_v3" is supported
            with pytest.raises(ValueError):
                highly_variable_genes(query, flavor="oopsie")  # type: ignore[arg-type]


@pytest.mark.experimental
@pytest.mark.live_corpus
def test_max_loess_jitter_error(small_mem_context: soma.SOMATileDBContext) -> None:
    with cellxgene_census.open_soma(census_version="stable", context=small_mem_context) as census:
        with pytest.raises(ValueError):
            get_highly_variable_genes(
                census,
                organism="mus_musculus",
                obs_value_filter='is_primary_data == True and tissue_general == "heart"',
                batch_key="cell_type",
                max_loess_jitter=0.0,
            )


@pytest.mark.experimental
@pytest.mark.live_corpus
@pytest.mark.parametrize(
    "batch_key",
    [
        None,
        "suspension_type",
        ["assay_ontology_term_id", "suspension_type"],
        ["dataset_id", "assay_ontology_term_id", "suspension_type", "donor_id"],
    ],
)
def test_hvg_user_defined_batch_key_func(
    small_mem_context: soma.SOMATileDBContext,
    batch_key: Union[None, str, list[str]],
) -> None:
    if batch_key is None:

        def batch_key_func(srs: pd.Series[Any]) -> str:
            raise AssertionError("should never be called without a batch key")

    else:
        if isinstance(batch_key, str):
            keys = set([batch_key])
        else:
            keys = set(batch_key)

        def batch_key_func(srs: pd.Series[Any]) -> str:
            assert set(srs.keys()) == keys
            return "batch0"

    with cellxgene_census.open_soma(census_version="latest", context=small_mem_context) as census:
        # Get the highly variable genes
        with census["census_data"]["mus_musculus"].axis_query(
            measurement_name="RNA",
            obs_query=soma.AxisQuery(coords=(slice(75000),)),
        ) as query:
            hvg = highly_variable_genes(
                query,
                batch_key=batch_key,
                batch_key_func=batch_key_func,
                n_top_genes=100,
            )

            assert len(hvg[hvg.highly_variable]) == 100



"""
An API to facilitate use of PyTorch ML training with data from the CZI Science CELLxGENE Census.
"""

from .pytorch import ExperimentDataPipe, Stats, experiment_dataloader

__all__ = [
    "Stats",
    "ExperimentDataPipe",
    "experiment_dataloader",
]



# Copyright (c) 2022-2023 Chan Zuckerberg Initiative Foundation
#
# Licensed under the MIT License.

"""
Experiments handler.

Contains methods to retrieve SOMA Experiments.
"""

import re

import tiledbsoma as soma


def _get_experiment(census: soma.Collection, organism: str) -> soma.Experiment:
    """Given a census ``soma.Collection``, return the experiment for the named organism.
    Organism matching is somewhat flexible, attempting to map from human-friendly
    names to the underlying collection element name.

    Args:
        census: soma.Collection
            The census.
        organism: str
            The organism name, eg., ``Homo sapiens``.

    Returns:
        A soma.Experiment object with the requested experiment.

    Raises:
        ValueError: if unable to find the specified organism.

    Lifecycle:
        maturing

    Examples:

        >>> human = get_experiment(census, 'homo sapiens')

        >>> human = get_experiment(census, 'Homo sapiens')

        >>> human = get_experiment(census, 'homo_sapiens')
    """
    # lower/snake case the organism name to find the experiment name
    exp_name = re.sub(r"[ ]+", "_", organism).lower()

    if exp_name not in census["census_data"]:
        raise ValueError(f"Unknown organism {organism} - does not exist")
    exp = census["census_data"][exp_name]
    if exp.soma_type != "SOMAExperiment":
        raise ValueError(f"Unknown organism {organism} - not a SOMA Experiment")

    return exp



# Copyright (c) 2022-2023 Chan Zuckerberg Initiative Foundation
#
# Licensed under the MIT License.

"""Get slice as AnnData

Methods to retrieve slices of the census as AnnData objects.
"""
from typing import Optional, Sequence

import anndata
import tiledbsoma as soma
from somacore.options import SparseDFCoord

from ._experiment import _get_experiment


def get_anndata(
    census: soma.Collection,
    organism: str,
    measurement_name: str = "RNA",
    X_name: str = "raw",
    X_layers: Optional[Sequence[str]] = (),
    obsm_layers: Optional[Sequence[str]] = (),
    obs_value_filter: Optional[str] = None,
    obs_coords: Optional[SparseDFCoord] = None,
    var_value_filter: Optional[str] = None,
    var_coords: Optional[SparseDFCoord] = None,
    column_names: Optional[soma.AxisColumnNames] = None,
) -> anndata.AnnData:
    """
    Convience wrapper around ``soma.Experiment`` query, to build and execute a query,
    and return it as an :class:`anndata.AnnData` object.

    Args:
        census:
            The census object, usually returned by :func:`cellxgene_census.open_soma()`.
        organism:
            The organism to query, usually one of `Homo sapiens` or `Mus musculus`.
        measurement_name:
            The measurement object to query. Defaults to `RNA`.
        X_name:
            The ``X`` layer to query. Defaults to `raw`.
        X_layers:
            Additional layers to add to ``anndata.AnnData.layers``.
        obs_value_filter:
            Value filter for the ``obs`` metadata. Value is a filter query written in the
            SOMA ``value_filter`` syntax.
        obs_coords:
            Coordinates for the ``obs`` axis, which is indexed by the ``soma_joinid`` value.
            May be an ``int``, a list of ``int``, or a slice. The default, ``None``, selects all.
        var_value_filter:
            Value filter for the ``var`` metadata. Value is a filter query written in the
            SOMA ``value_filter`` syntax.
        var_coords:
            Coordinates for the ``var`` axis, which is indexed by the ``soma_joinid`` value.
            May be an ``int``, a list of ``int``, or a slice. The default, ``None``, selects all.
        column_names:
            Columns to fetch for ``obs`` and ``var`` dataframes.
        obsm_layers:
            Additional obsm layers to read and return in the ``obsm`` slot.

    Returns:
        An :class:`anndata.AnnData` object containing the census slice.

    Lifecycle:
        maturing

    Examples:
        >>> get_anndata(census, "Mus musculus", obs_value_filter="tissue_general in ['brain', 'lung']")

        >>> get_anndata(census, "Homo sapiens", column_names={"obs": ["tissue"]})

        >>> get_anndata(census, "Homo sapiens", obs_coords=slice(0, 1000))
    """
    exp = _get_experiment(census, organism)
    obs_coords = (slice(None),) if obs_coords is None else (obs_coords,)
    var_coords = (slice(None),) if var_coords is None else (var_coords,)
    with exp.axis_query(
        measurement_name,
        obs_query=soma.AxisQuery(value_filter=obs_value_filter, coords=obs_coords),
        var_query=soma.AxisQuery(value_filter=var_value_filter, coords=var_coords),
    ) as query:
        return query.to_anndata(X_name=X_name, column_names=column_names, X_layers=X_layers, obsm_layers=obsm_layers)



"""
An API to facilitate using Hugging Face ML tools with the CZI Science CELLxGENE Census.
"""

from .cell_dataset_builder import CellDatasetBuilder
from .geneformer_tokenizer import GeneformerTokenizer

__all__ = [
    "CellDatasetBuilder",
    "GeneformerTokenizer",
]
